
import os as _os
import uuid as _uuid

__all__ = ["OCIAccount"]


class OCIAccount:
    """This class abstracts all interaction with OCI login accounts. This
       is a low-level account that allows us to connect to the object
       store at a low-level and to call OCI functions
    """
    @staticmethod
    def _assert_valid_login_dict(login):
        """This function validates that the passed login dictionary
           contains all of the keys needed to login"""

        if login is None:
            from Acquire.Service import AccountError
            raise AccountError("You need to supply login credentials!")

        if not isinstance(login, dict):
            from Acquire.Service import AccountError
            raise AccountError(
                "You need to supply a valid login credential dictionary!")

        needed_keys = ["user", "key_lines", "fingerprint",
                       "tenancy", "region", "pass_phrase"]

        missing_keys = []

        for key in needed_keys:
            if key not in login:
                missing_keys.append(key)

        if len(missing_keys) > 0:
            from Acquire.Service import AccountError
            raise AccountError(
                "Cannot log in as the login dictionary "
                "is missing the following data: %s" % str(missing_keys))

    @staticmethod
    def get_login(login):
        """This function turns the passed login details into
           a valid oci login
        """

        # validate that all of the information is held in the
        # 'login' dictionary
        OCIAccount._assert_valid_login_dict(login)

        # first, we need to convert the 'login' so that it puts
        # the private key into a file
        keyfile = "/tmp/key.pem"

        try:
            with open(keyfile, "w") as FILE:
                for line in login["key_lines"]:
                    FILE.write(line)

            _os.chmod(keyfile, 0o0600)

            del login["key_lines"]
            login["key_file"] = keyfile
        except:
            _os.remove(keyfile)
            raise

        return login

    @staticmethod
    def _sanitise_bucket_name(bucket_name):
        """This function sanitises the passed bucket name. It will always
           return a valid bucket name. If "None" is passed, then a new,
           unique bucket name will be generated"""

        if bucket_name is None:
            return str(_uuid.uuid4())

        return "_".join(bucket_name.split())

    @staticmethod
    def create_and_connect_to_bucket(login_details, compartment,
                                     bucket_name=None):
        """Connect to the object store compartment 'compartment'
           using the passed 'login_details', and create a bucket
           called 'bucket_name". Return a handle to the
           created bucket. If the bucket already exists this will return
           a handle to the existing bucket
        """
        try:
            from oci.object_storage import ObjectStorageClient as \
                _ObjectStorageClient
            from oci.object_storage.models import CreateBucketDetails as \
                _CreateBucketDetails
        except:
            raise ImportError(
                "Cannot import OCI. Please install OCI, e.g. via "
                "'pip install oci' so that you can connect to the "
                "Oracle Cloud Infrastructure")

        login = OCIAccount.get_login(login_details)
        bucket = {}
        client = None

        try:
            client = _ObjectStorageClient(login)

            bucket["client"] = client
            bucket["compartment_id"] = compartment

            # save the region as this is needed for some services
            bucket["region"] = login["region"]

            namespace = client.get_namespace().data
            bucket["namespace"] = namespace
            bucket["bucket_name"] = bucket_name

            try:
                request = _CreateBucketDetails()
                request.compartment_id = compartment
                request.name = OCIAccount._sanitise_bucket_name(bucket_name)

                bucket["bucket"] = client.create_bucket(
                                            client.get_namespace().data,
                                            request).data
            except Exception as e1:
                # couldn't create the bucket - likely because it already
                # exists - try to connect to the existing bucket
                try:
                    bucket["bucket"] = client.get_bucket(namespace,
                                                         bucket_name).data
                except Exception as e:
                    from Acquire.Service import AccountError
                    raise AccountError(
                        "Cannot access the bucket '%s' : %s (originally %s)" %
                        (bucket_name, str(e), str(e1)))
        except:
            _os.remove(_os.path.abspath(login["key_file"]))
            raise

        _os.remove(_os.path.abspath(login["key_file"]))

        return bucket

    @staticmethod
    def connect_to_bucket(login_details, compartment, bucket_name):
        """Connect to the object store compartment 'compartment'
           using the passed 'login_details', returning a handle to the
           bucket associated with 'bucket
        '"""
        try:
            from oci.object_storage import ObjectStorageClient as \
                _ObjectStorageClient
        except:
            raise ImportError(
                "Cannot import OCI. Please install OCI, e.g. via "
                "'pip install oci' so that you can connect to the "
                "Oracle Cloud Infrastructure")

        login = OCIAccount.get_login(login_details)
        bucket = {}

        try:
            client = _ObjectStorageClient(login)
            bucket["client"] = client
            bucket["compartment_id"] = compartment

            namespace = client.get_namespace().data
            bucket["namespace"] = namespace

            bucket["bucket"] = client.get_bucket(namespace, bucket_name).data
            bucket["bucket_name"] = bucket_name
        except:
            _os.remove(_os.path.abspath(login["key_file"]))
            raise

        _os.remove(_os.path.abspath(login["key_file"]))

        return bucket




import uuid as _uuid
from copy import copy as _copy

from Acquire.Service import Service as _Service

__all__ = ["AccessService"]


class AccessService(_Service):
    """This is a specialisation of Service for Access Services"""
    def __init__(self, other=None):
        if isinstance(other, _Service):
            self.__dict__ = _copy(other.__dict__)

            if not self.is_access_service():
                from Acquire.Access import AccessServiceError
                raise AccessServiceError(
                    "Cannot construct an AccessService from "
                    "a service which is not an access service!")
        else:
            _Service.__init__(self)

    def _call_local_function(self, function, args):
        """Internal function called to short-cut local 'remote'
           function calls
        """
        from access.route import access_functions as _access_functions
        from admin.handler import create_handler as _create_handler
        handler = _create_handler(_access_functions)
        return handler(function, args)

    def get_trusted_storage_service(self):
        """Return a trusted storage service"""
        from Acquire.Service import get_trusted_services \
            as _get_trusted_services

        services = _get_trusted_services()

        try:
            return services["storage"][0]
        except:
            from Acquire.Service import ServiceError
            raise ServiceError(
                "There is no trusted storage service known to this access "
                "service")




import datetime as _datetime
import uuid as _uuid

import base64 as _base64

__all__ = ["LoginSession"]


class LoginSession:
    """This class holds all details of a single login session"""
    def __init__(self, public_key=None, public_cert=None, ip_addr=None,
                 hostname=None, login_message=None):
        self._pubkey = None
        self._uid = None
        self._request_datetime = None
        self._login_datetime = None
        self._logout_datetime = None
        self._pubcert = None
        self._status = None
        self._ipaddr = None
        self._hostname = None
        self._login_message = None

        from Acquire.Crypto import PublicKey as _PublicKey

        if public_key:
            try:
                public_key = open(public_key, "r").readlines()
            except:
                pass

            try:
                self._pubkey = _PublicKey.from_data(public_key)
            except:
                self._pubkey = public_key

            if not isinstance(self._pubkey, _PublicKey):
                raise TypeError("The public key must be of type PublicKey")

            from Acquire.ObjectStore import get_datetime_now \
                as _get_datetime_now

            self._uid = str(_uuid.uuid4())
            self._request_datetime = _get_datetime_now()
            self._status = "unapproved"

        if public_cert:
            try:
                public_cert = open(public_cert, "r").readlines()
            except:
                pass

            try:
                self._pubcert = _PublicKey.from_data(public_cert)
            except:
                self._pubcert = public_cert

            if not isinstance(self._pubcert, _PublicKey):
                raise TypeError("The public certificate must be of "
                                "type PublicKey")

        if ip_addr:
            self._ipaddr = ip_addr

        if hostname:
            self._hostname = hostname

        if login_message:
            self._login_message = login_message

    def __str__(self):
        if self.is_null():
            return "LoginSession::null"
        else:
            return "LoginSession(uid=%s, status=%s)" % \
                            (self.uid(), self.status())

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self._uid == other._uid
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def is_null(self):
        """Return whether or not this object is null"""
        return self._uid is None

    def public_key(self):
        """Return the public key"""
        if self._pubkey is None:
            from Acquire.Identity import LoginSessionError
            raise LoginSessionError(
                "You cannot get a public key from "
                "a logged out or otherwise denied session")

        elif self.is_suspicious():
            from Acquire.Identity import LoginSessionError
            raise LoginSessionError(
                "You cannot get a public key from "
                "a login session that has been marked as suspicious")

        return self._pubkey

    def public_certificate(self):
        """Return the public certificate"""
        if self._pubcert is None:
            from Acquire.Identity import LoginSessionError
            raise LoginSessionError(
               "You cannot get a public certificate from "
               "a logged out or otherwise denied session")

        elif self.is_suspicious():
            from Acquire.Identity import LoginSessionError
            raise LoginSessionError(
                "You cannot get a public certificate from "
                "a login session that has been marked as suspicious")

        return self._pubcert

    def request_source(self):
        """Return the IP address of the source of
           this request. This is used to rate limit someone
           who is maliciously requesting logins..."""
        return self._ipaddr

    def uuid(self):
        """Return the UUID of this request"""
        return self._uid

    @staticmethod
    def to_short_uid(long_uid):
        """Return the short UID version of the passed long uid"""
        return long_uid[:8]

    def short_uuid(self):
        """Return a short UUID that will be used to
           provide a more human-readable session ID"""
        if self._uid:
            return LoginSession.to_short_uid(self._uid)
        else:
            return None

    def uid(self):
        """Synonym for uuid"""
        return self.uuid()

    def short_uid(self):
        """Synonym for short_uuid"""
        return self.short_uuid()

    def regenerate_uuid(self):
        """Regenerate the UUID as there has been a clash"""
        if self._pubkey:
            self._uid = str(_uuid.uuid4())

    def regenerate_uid(self):
        """Synonym for regenerate_uuid"""
        self.regenerate_uuid()

    def creation_time(self):
        """Return the date and time when this was created"""
        return self._request_datetime

    def login_time(self):
        """Return the date and time when the user logged in. This
           returns None if the user has not yet logged in"""
        return self._login_datetime

    def logout_time(self):
        """Return the date and time when the user logged out. This
           returns None if the user has not yet logged out"""
        return self._logout_datetime

    def hours_since_creation(self):
        """Return the number of hours since this request was
           created. This will return a float, so 1 second is
           1 / 3600th of an hour"""

        if self._request_datetime:
            from Acquire.ObjectStore import get_datetime_now \
                as _get_datetime_now
            delta = _get_datetime_now() - self._request_datetime
            return delta.total_seconds() / 3600.0
        else:
            return 0

    def status(self):
        """Return the status of this login session"""
        return self._status

    def set_suspicious(self):
        """Put this login session into a suspicious state. This
           will be because weird activity has been detected which indicates
           that the session may be have been cracked. A login session
           in a suspicious state should not be granted any permissions.
        """
        if self._uid:
            self._status = "suspicious"

    def set_approved(self):
        """Register that this request has been approved"""
        if self._uid:
            if self._pubkey is None or self._pubcert is None:
                from Acquire.Identity import LoginSessionError
                raise LoginSessionError(
                    "You cannot approve a login session "
                    "that doesn't have a valid public key or certificate. "
                    "Perhaps you have already denied the authorisation "
                    "or logged out?")

            if self.status() != "unapproved":
                from Acquire.Identity import LoginSessionError
                raise LoginSessionError(
                    "You cannot approve a login session "
                    "that is not in the 'unapproved' state. This login "
                    "session is in the '%s' state." % self.status())

            from Acquire.ObjectStore import get_datetime_now \
                as _get_datetime_now
            self._login_datetime = _get_datetime_now()
            self._status = "approved"

    def _clear_keys(self):
        """Function called to remove all keys
           from this session, as it has now been terminated
           (and so the keys are no longer valid)
        """
        self._pubkey = None

    def set_denied(self):
        """Register that this request has been denied"""
        if self._uid:
            self._status = "denied"
            self._clear_keys()
            # also clear the certificate
            self._pubcert = None

    def set_logged_out(self):
        """Register that this request has been closed as
           the user has logged out"""
        if self._uid:
            from Acquire.ObjectStore import get_datetime_now \
                as _get_datetime_now

            self._status = "logged_out"
            self._logout_datetime = _get_datetime_now()
            self._clear_keys()

    def login(self):
        """Convenience function to set the session into the logged in state"""
        self.set_approved()

    def logout(self):
        """Convenience function to set the session into the logged out state"""
        self.set_logged_out()

    def is_suspicious(self):
        """Return whether or not this session is suspicious"""
        if self._status:
            return self._status == "suspicious"
        else:
            return False

    def is_approved(self):
        """Return whether or not this session is open and
           approved by the user"""
        if self._status:
            return self._status == "approved"
        else:
            return False

    def is_logged_out(self):
        """Return whether or not this session has logged out"""
        if self._status:
            return self._status == "logged_out"
        else:
            return False

    def login_message(self):
        """Return any login message that has been set by the user"""
        return self._login_message

    def hostname(self):
        """Return the user-supplied hostname of the host making the
           login request"""
        return self._hostname

    def to_data(self):
        """Return a data version (dictionary) of this LoginSession"""

        if self._uid is None:
            return None

        from Acquire.ObjectStore import datetime_to_string \
            as _datetime_to_string

        data = {}
        data["uid"] = self._uid
        data["request_datetime"] = _datetime_to_string(self._request_datetime)

        try:
            data["login_datetime"] = _datetime_to_string(self._login_datetime)
        except:
            data["login_datetime"] = None

        try:
            data["logout_datetime"] = _datetime_to_string(
                                            self._logout_datetime)
        except:
            data["logout_datetime"] = None

        if self._pubkey:
            data["public_key"] = self._pubkey.to_data()

        if self._pubcert:
            data["public_certificate"] = self._pubcert.to_data()

        data["status"] = self._status
        data["ipaddr"] = self._ipaddr
        data["hostname"] = self._hostname
        data["login_message"] = self._login_message

        return data

    @staticmethod
    def from_data(data):
        """Return a LoginSession constructed from the passed data
           (dictionary)
        """
        if data is None:
            return None

        try:
            from Acquire.ObjectStore import string_to_datetime \
                as _string_to_datetime
            from Acquire.Crypto import PublicKey as _PublicKey

            logses = LoginSession()

            try:
                logses._uid = data["uid"]
            except:
                logses._uid = data["uuid"]  # for backward compatibility

            logses._request_datetime = _string_to_datetime(
                                            data["request_datetime"])

            try:
                logses._login_datetime = _string_to_datetime(
                                            data["login_datetime"])
            except:
                logses._login_datetime = None

            try:
                logses._logout_datetime = _string_to_datetime(
                                            data["logout_datetime"])
            except:
                logses._logout_datetime = None

            try:
                logses._pubkey = _PublicKey.from_data(data["public_key"])
            except:
                logses._pubkey = None

            try:
                logses._pubcert = _PublicKey.from_data(
                                            data["public_certificate"])
            except:
                logses._pubcert = None

            logses._status = data["status"]
            logses._ipaddr = data["ipaddr"]

            logses._hostname = data["hostname"]
            logses._login_message = data["login_message"]

            return logses

        except Exception as e:
            from Acquire.Service import exception_to_string
            from Acquire.Identity import LoginSessionError
            raise LoginSessionError(
                "Cannot load the LoginSession from "
                "the object store?\n\nCAUSE: %s" % (exception_to_string(e)))



#!/bin/env python3


def main():
    import argparse
    import sys

    from Acquire.Client import Wallet, LoginError

    parser = argparse.ArgumentParser(
                    description="Log into an Acquire-based identity "
                                "service via a login url",
                    prog="acquire_login")

    parser.add_argument("url", type=str, nargs="*",
                        help="Login URL")

    parser.add_argument('-u', '--username', type=str, nargs='?',
                        help="Username with which to log in")

    parser.add_argument('--remember-password', action="store_true",
                        default=True,
                        help="Remember the password (default on)")

    parser.add_argument('--remember-device', action="store_true",
                        default=None,
                        help="Remember this device (saves OTP code, "
                             "default off)")

    parser.add_argument('--no-remember-device', action="store_true",
                        default=None,
                        help="Don't remember this device, and don't ask to")

    parser.add_argument('--no-remember-password', action="store_true",
                        default=None,
                        help="Don't remember the password, and don't ask to")

    parser.add_argument('--remove', type=str, nargs="*",
                        help="Remove locally stored information about the "
                        "passed account(s)")

    parser.add_argument('--remove-service', type=str, nargs="*",
                        help="Remove locally stored information about the "
                        "passed service(s)")

    parser.add_argument('--dry-run', action="store_true", default=None,
                        help="Do a dry-run of the login - don't connect to "
                        "the server")

    args = parser.parse_args()

    remember_device = args.remember_device

    if args.no_remember_device:
        remember_device = False

    remember_password = args.remember_password

    if remember_password is None:
        remember_password = True

    if args.no_remember_password:
        remember_password = False

    dryrun = args.dry_run

    if not remember_password:
        # should not remember the otpsecret if
        # we don't trust this to remember the password!
        remember_device = False

    do_nothing = True

    if args.remove:
        for username in args.remove:
            try:
                do_nothing = False
                print("Removing locally stored credentials for '%s'"
                      % username)
                Wallet.remove_user_info(username)
            except Exception as e:
                print(e)
                pass

    if args.remove_service:
        for service in args.remove_service:
            try:
                do_nothing = False
                print("Removing locally stored information "
                      "about service '%s'" % service)
                Wallet.remove_service(service)
            except Exception as e:
                print(e)
                pass

    if do_nothing and len(args.url) == 0:
        parser.print_help(sys.stdout)

    if len(args.url) == 0:
        sys.exit(0)

    wallet = Wallet()

    for url in args.url:
        try:
            wallet.send_password(url, args.username,
                                 remember_password,
                                 remember_device,
                                 dryrun)
        except LoginError as e:
            print("\n%s" % e.args)
        except Exception as e:
            from Acquire.Service import exception_to_string
            print(exception_to_string(e))

if __name__ == "__main__":
    main()




import datetime as _datetime
import json as _json

from ._errors import PaymentError

__all__ = ["Cheque"]


class Cheque:
    """This class acts like a real world cheque, except it can only
       be used to transfer money from a regular user to a service
       user's account. This allows users to pay for services. This can be
       written by a user against one of their accounts, to be sent to a
       recipient service to pay for a named service. The recipient can
       send the cheque to the accounting service to trigger payment,
       upon which a CreditNote will be returned. Once receipted,
       payment will be complete.
    """
    def __init__(self):
        self._cheque = None
        self._accounting_service_url = None

    @staticmethod
    def write(account=None, resource=None,
              recipient_url=None, max_spend=None,
              expiry_date=None):
        """Create and return a cheque that can be used at any point
           in the future to authorise a transaction. If 'canonical_url'
           is supplied, then only the service with matching canonical
           url can 'cash' the cheque (it will need to sign the cheque
           before sending it to the accounting service). If 'max_spend'
           is specified, then the cheque is only valid up to that
           maximum spend. Otherwise, it is valid up to the maximum
           daily spend limit (or other limits) of the account. If
           'expiry_date' is supplied then this cheque is valid only
           before the supplied datetime. If 'resource' is
           supplied then this cheque is only valid to pay for the
           specified resource (this should be a string that everyone
           agrees represents the resource in question). Note that
           this cheque is for a future transaction. We do not check
           to see if there are sufficient funds now, and this does
           not affect the account. If there are insufficient funds
           when the cheque is cashed (or it breaks spending limits)
           then the cheque will bounce.
        """
        from Acquire.Client import Account as _Account

        if not isinstance(account, _Account):
            raise TypeError("You must pass a valid Acquire.Client.Account "
                            "object to write a cheque...")

        if max_spend is not None:
            from Acquire.ObjectStore import decimal_to_string \
                as _decimal_to_string
            max_spend = _decimal_to_string(max_spend)

        if expiry_date is not None:
            from Acquire.ObjectStore import datetime_to_string \
                as _datetime_to_string
            expiry_date = _datetime_to_string(expiry_date)

        if recipient_url is not None:
            recipient_url = str(recipient_url)

        from Acquire.ObjectStore import create_uuid as _create_uuid
        from Acquire.Identity import Authorisation as _Authorisation

        info = _json.dumps({"recipient_url": recipient_url,
                            "max_spend": max_spend,
                            "expiry_date": expiry_date,
                            "uid": _create_uuid(),
                            "resource": str(resource),
                            "account_uid": account.uid()})

        auth = _Authorisation(user=account.user(), resource=info)

        data = {"info": info, "authorisation": auth.to_data()}

        cheque = Cheque()

        cheque._cheque = account.accounting_service().encrypt_data(data)
        cheque._accounting_service_url = \
            account.accounting_service().canonical_url()

        return cheque

    def read(self, spend, resource, receipt_by):
        """Read the cheque - this will read the cheque to return the
           decrypted contents. This will only work if this function
           is called on the accounting service that will cash the
           cheque, if the signature on the cheque matches the
           service that is authorised to cash the cheque, and
           if the passed resource matches the resource
           encoded in the cheque. If this is all correct, then the
           returned dictionary will contain;

           {"recipient_url": The URL of the service which was sent the cheque,
            "recipient_key_fingerprint": Verified fingerprint of the service
                                         key that signed this cheque
            "spend": The amount authorised by this cheque,
            "uid": The unique ID for this cheque,
            "resource": String that identifies the resource this cheque will
                        be used to pay for,
            "account_uid": UID of the account from which funds will be drawn
            "authorisation" : Verified authorisation from the user who
                              says they own the account for the spend
            "receipt_by" : Time when we must receipt the cheque, or
                           we will lose the money
           }

           You must pass in the spend you want to draw from the cheque,
           a string representing the resource this cheque will
           be used to pay for, and the time by which you promise to receipt
           the cheque after cashing
        """
        if self._cheque is None:
            raise PaymentError("You cannot read a null cheque")

        from Acquire.ObjectStore import string_to_decimal \
            as _string_to_decimal
        from Acquire.ObjectStore import string_to_datetime \
            as _string_to_datetime
        from Acquire.ObjectStore import datetime_to_string \
            as _datetime_to_string
        from Acquire.Service import get_this_service as _get_this_service

        spend = _string_to_decimal(spend)
        resource = str(resource)
        receipt_by = _string_to_datetime(receipt_by)

        service = _get_this_service(need_private_access=True)

        # get the cheque data - this may have been signed
        try:
            cheque_data = _json.loads(self._cheque["signed_data"])
        except:
            cheque_data = self._cheque

        # decrypt the cheque's data - only possible on the accounting service
        cheque_data = service.decrypt_data(cheque_data)

        # the date comprises the user-authorisation that acts as a
        # signature that the user wrote this cheque, and the info
        # for the cheque to say how it is valid
        from Acquire.Identity import Authorisation as _Authorisation
        auth = _Authorisation.from_data(cheque_data["authorisation"])
        info = cheque_data["info"]

        # validate that the user authorised this cheque
        try:
            auth.verify(resource=info)
        except Exception as e:
            raise PaymentError(
                "The user's signature/authorisation for this cheque "
                "is not valid! ERROR: %s" % str(e))

        info = _json.loads(info)

        # the user signed this cheque :-)
        info["authorisation"] = auth

        # check the signature if one was needed
        try:
            recipient_url = info["recipient_url"]
        except:
            recipient_url = None

        if recipient_url:
            # the recipient was specified - verify that we trust
            # the recipient, and that they have signed the cheque
            recipient_service = service.get_trusted_service(
                                            service_url=recipient_url)
            recipient_service.verify_data(self._cheque)
            info["recipient_key_fingerprint"] = self._cheque["fingerprint"]

        # validate that the item signature is correct
        try:
            cheque_resource = info["resource"]
        except:
            cheque_resource = None

        if cheque_resource is not None:
            if resource != resource:
                raise PaymentError(
                    "Disagreement over the resource for which "
                    "this cheque has been signed")

        info["resource"] = resource

        try:
            max_spend = info["max_spend"]
            del info["max_spend"]
        except:
            max_spend = None

        if max_spend is not None:
            max_spend = _string_to_decimal(max_spend)

            if max_spend < spend:
                raise PaymentError(
                    "The requested spend (%s) exceeds the authorised "
                    "maximum value of the cheque" % (spend))

        info["spend"] = spend

        try:
            expiry_date = info["expiry_date"]
            del expiry_date["expiry_date"]
        except:
            expiry_date = None

        if expiry_date is not None:
            expiry_date = _string_to_datetime(expiry_date)

            # validate that the cheque will not have expired
            # when we receipt it
            from Acquire.ObjectStore import get_datetime_now \
                as _get_datetime_now
            now = _get_datetime_now()

            if now > receipt_by:
                raise PaymentError(
                    "The time when you promised to receipt the cheque "
                    "has already passed!")

            if receipt_by > expiry_date:
                raise PaymentError(
                    "The cheque will have expired after you plan to "
                    "receipt it!: %s versus %s" %
                    (_datetime_to_string(receipt_by),
                     _datetime_to_string(expiry_date)))

        info["receipt_by"] = receipt_by

        print(info)

        # everything now checks out - return the read cheque
        return info

    def cash(self, spend, resource, receipt_within=3600):
        """Cash this cheque, specifying how much to be cashed,
           and the resource that will be paid for
           using this cheque. This will send the cheque to the
           accounting service (if we trust that accounting service).
           The accounting service will check that the cheque is valid,
           and the signature of the item is correct. It will then
           withdraw 'spend' funds from the account that signed the
           cheque, returning valid CreditNote(s) that can be trusted
           to show that the funds exist.

           If 'receipt_within' is set, then the CreditNotes will
           be automatically cancelled if they are not
           receipted within 'receipt_within' seconds

           It is your resposibility to receipt the note for
           the actual valid incurred once the service has been
           delivered, thereby actually transferring the cheque
           funds into your account (on that accounting service)

           This returns a list of the CreditNote(s) that were
           cashed from the cheque
        """
        if self._cheque is None:
            raise PaymentError("You cannot cash a null cheque!")

        # sign the cheque to show we have seen it
        from Acquire.Service import get_this_service as _get_this_service
        service = _get_this_service(need_private_access=True)
        self._cheque = service.sign_data(self._cheque)

        # get the trusted accounting service that will honour the cheque
        accounting_service = self.accounting_service()

        # when do we guarantee to receipt the credit notes by?
        from Acquire.ObjectStore import get_datetime_future \
            as _get_datetime_future
        receipt_by = _get_datetime_future(receipt_within)

        # which account should the money be paid into?
        account_uid = service.service_user_account_uid(
                                accounting_service=accounting_service)

        # next - send the cheque to the accounting service to
        # show that we know the item_id and want to cash it
        from Acquire.ObjectStore import decimal_to_string \
            as _decimal_to_string
        from Acquire.ObjectStore import datetime_to_string \
            as _datetime_to_string
        from Acquire.ObjectStore import string_to_list \
            as _string_to_list

        result = accounting_service.call_function(
            function="cash_cheque",
            args={"cheque": self.to_data(),
                  "spend": _decimal_to_string(spend),
                  "resource": str(resource),
                  "account_uid": account_uid,
                  "receipt_by": _datetime_to_string(receipt_by)})

        credit_notes = None

        try:
            from Acquire.Accounting import CreditNote as _CreditNote
            credit_notes = _string_to_list(result["credit_notes"],
                                           _CreditNote)
        except Exception as e:
            raise PaymentError(
                "Attempt to cash the cheque has not resulted in a "
                "valid CreditNote? Error = %s" % str(e))

        total_cashed = 0

        for note in credit_notes:
            total_cashed = total_cashed + note.value()
            if note.account_uid() != account_uid:
                raise PaymentError(
                    "The cashed cheque is paying into the wrong account! "
                    "%s. It should be going to %s" %
                    (note.account_uid(), account_uid))

        if total_cashed != spend:
            raise PaymentError(
                "The value of the cheque (%s) does not match the total value "
                "of the credit note(s) returned (%s)" % (spend, total_cashed))

        return credit_notes

    def accounting_service_url(self):
        """Return the URL of the accounting service that will honour
           this cheque
        """
        return self._accounting_service_url

    def accounting_service(self):
        """Return the accounting service that will honour this cheque.
           Note that this will only return the service if it is trusted
           by the service on which this function is called
        """
        from Acquire.Service import get_this_service as _get_this_service
        service = _get_this_service()
        accounting_service = service.get_trusted_service(
                                            self.accounting_service_url())

        if not accounting_service.is_accounting_service():
            from Acquire.Service import ServiceError
            raise ServiceError(
                "The service that is supposed to honour the cheque (%s) "
                "does not appear to be a valid accounting service" %
                (str(accounting_service)))

        return accounting_service

    def to_data(self):
        """Return a json-serialisable dictionary for this object"""
        data = {}

        if self._cheque is not None:
            data["accounting_service_url"] = self._accounting_service_url
            data["cheque"] = self._cheque

        return data

    @staticmethod
    def from_data(data):
        """Return a cheque constructed from the passed (json-deserialised
           dictionary)"""
        cheque = Cheque()

        if (data and len(data) > 0):
            cheque._cheque = data["cheque"]
            cheque._accounting_service_url = data["accounting_service_url"]

        return cheque




from cachetools import LRUCache as _LRUCache
from cachetools import cached as _cached

__all__ = ["Service"]

_cache_service_lookup = _LRUCache(maxsize=5)


@_cached(_cache_service_lookup)
def _get_remote_service(service_url):
    """Call this function to get the Service info of a remote Service.
       This will check if we have seen the service before. If so, then
       we will ensure that this is the correct service. If not, then
       we will prompt the user to validate that this is the right
       service
    """
    from Acquire.Client import Wallet as _Wallet
    return _Wallet.get_service(service_url)


class Service:
    """This class represents a service in the system. Services
       will either be identity services, access services,
       storage services or accounting services.

       This class provides a client-side wrapper to fetch or set-up
       a service. This is a metamorphic (shape-shifting) class,
       i.e. during construction it will transform into the class
       of the type of service, e.g. Acquire.Identity.IdentityService
    """
    def __init__(self, service_url=None, service_uid=None):
        """Construct the service that is accessed at the remote
           URL 'service_url'. This will fetch and return the
           details of the remote service. This wrapper is a
           chameleon class, and will transform into the
           class type of the fetched service, e.g.

            service = Acquire.Client.Service("https://identity_service_url")
            service.__class__ == Acquire.Identity.IdentityService
        """
        try:
            from Acquire.Client import Wallet as _Wallet
            service = _Wallet.get_service(service_url)

            from copy import copy as _copy
            self.__dict__ = _copy(service.__dict__)
            self.__class__ = service.__class__
        except Exception as e:
            _cache_service_lookup.clear()
            self._failed = True
            raise e

    def _fail(self):
        """This is called by all functions as this Service
           has failed to be initialised
        """
        if self._failed:
            from Acquire.Service import ServiceError
            raise ServiceError(
                "Cannot do anything with a null service")

    def __str__(self):
        return "Service(failed setup!)"

    def uuid(self):
        """Synonym for uid"""
        return self.uid()

    def uid(self):
        """Return the uuid of this service. This MUST NEVER change, as
           the UID uniquely identifies this service to all other
           services
        """
        self._fail()
        return None  # pylint: disable=W1111

    def service_type(self):
        """Return the type of this service"""
        self._fail()
        return None

    def is_locked(self):
        """Return whether or not this service object is locked. Locked
           service objects don't contain copies of any private keys,
           and can be safely shared as a means of distributing public
           keys and certificates
        """
        self._fail()
        return None

    def is_unlocked(self):
        """Return whether or not this service object is unlocked. Unlocked
           service objects have access to the skeleton key and other private
           keys. They should only run on the service. Locked service objects
           are what are returned by services to provide public keys and
           public certificates
        """
        self._fail()
        return None

    def get_trusted_service(self, service_url=None, service_uid=None):
        """Return the trusted service info for the service with specified
           service_url or service_uid"""
        self._fail()
        return None

    def assert_unlocked(self):
        """Assert that this service object is unlocked"""
        self._fail()
        return None

    def assert_admin_authorised(self, authorisation, resource=None):
        """Validate that the passed authorisation is valid for the
           (optionally) specified resource, and that this has been
           authorised by one of the admin accounts of this service
        """
        self._fail()
        return None

    def last_key_update(self):
        """Return the datetime when the key and certificate of this
           service were last updated
        """
        self._fail()
        return None

    def key_update_interval(self):
        """Return the time delta between server key updates"""
        self._fail()
        return None

    def should_refresh_keys(self):
        """Return whether the keys and certificates need to be refreshed
           - i.e. more than 'key_update_interval' has passed since the last
           key update
        """
        self._fail()
        return None

    def refresh_keys(self):
        """Refresh the keys and certificates"""
        self._fail()
        return None

    def can_identify_users(self):
        """Return whether or not this service can identify users.
           Most services can, at a minimum, identify their admin
           users. However, only true Identity Services can register
           and manage normal users
        """
        self._fail()
        return None

    def is_identity_service(self):
        """Return whether or not this is an identity service"""
        self._fail()
        return None

    def is_access_service(self):
        """Return whether or not this is an access service"""
        self._fail()
        return None

    def is_accounting_service(self):
        """Return whether or not this is an accounting service"""
        self._fail()
        return None

    def is_storage_service(self):
        """Return whether or not this is a storage service"""
        self._fail()
        return None

    def service_url(self):
        """Return the URL used to access this service"""
        self._fail()
        return None

    def canonical_url(self):
        """Return the canonical URL for this service (this is the URL the
           service thinks it has, and which it has used to register itself
           with all other services)
        """
        self._fail()
        return None

    def hostname(self):
        """Return the hostname of the canonical URL that provides
           this service
        """
        self._fail()
        return None

    def uses_https(self):
        """Return whether or not the canonical URL of this service
           is connected to via https
        """
        self._fail()
        return None

    def update_service_url(self, service_url):
        """Update the service url to be 'service_url'"""
        self._fail()
        return None

    def service_user_uid(self):
        """Return the UID of the service user account for this service"""
        self._fail()
        return None

    def service_user_name(self):
        """Return the name of the service user account for this service"""
        self._fail()
        return None

    def service_user_secrets(self):
        """Return the (encrypted) secrets for the service user account.
           These will only be returned if you have unlocked this service.
           You need access to the skeleton key to decrypt these secrets
        """
        self._fail()
        return None

    def login_service_user(self):
        """Return a logged in Acquire.Client.User for the service user.
           This can only be called inside the service, and when you
           have unlocked this service object
        """
        self._fail()
        return None

    def service_user_account_uid(self, accounting_service_url=None,
                                 accounting_service=None):
        """Return the UID of the financial account associated with
           this service on the passed accounting service
        """
        self._fail()
        return None

    def skeleton_key(self):
        """Return the skeleton key used by this service. This is an
           unchanging key which is stored internally, should never be
           shared outside the service, and which is used to encrypt
           all data. Unlocking the service involves loading and
           decrypting this skeleton key
        """
        self._fail()
        return None

    def private_key(self):
        """Return the private key (if it has been unlocked)"""
        self._fail()
        return None

    def private_certificate(self):
        """Return the private signing certificate (if it has been unlocked)"""
        self._fail()
        return None

    def public_key(self):
        """Return the public key for this service"""
        self._fail()
        return None

    def public_certificate(self):
        """Return the public signing certificate for this service"""
        self._fail()
        return None

    def last_key(self):
        """Return the old private key for this service (if it has
           been unlocked). This was the key used before the last
           key update, and we store it in case we have to decrypt
           data that was recently encrypted using the old public key
        """
        self._fail()
        return None

    def last_certificate(self):
        """Return the old public certificate for this service. This was the
           certificate used before the last key update, and we store it
           in case we need to verify data signed using the old private
           certificate
        """
        self._fail()
        return None

    def call_function(self, function, args=None):
        """Call the function 'func' on this service, optionally passing
           in the arguments 'args'. This is a simple wrapper around
           Acquire.Service.call_function which automatically
           gets the correct URL, encrypts the arguments using the
           service's public key, and supplies a key to encrypt
           the response (and automatically then decrypts the
           response)
        """
        self._fail()
        return {}

    def sign(self, message):
        """Sign the specified message"""
        self._fail()
        return None

    def verify(self, signature, message):
        """Verify that this service signed the message"""
        self._fail()
        return None

    def encrypt(self, message):
        """Encrypt the passed message"""
        self._fail()
        return None

    def decrypt(self, message):
        """Decrypt the passed message"""
        self._fail()
        return None

    def sign_data(self, data):
        """Sign the passed data, ready for transport. Data should be
           a json-serialisable dictionary. This will return a new
           json-serialisable dictionary, which will contain the
           signature and json-serialised original data, e.g. as;

           data = {"service_uid" : "SERVICE_UID",
                   "fingerprint" : "KEY_FINGERPRINT",
                   "signed_data" : "JSON_ENCODED_DATA",
                   "signature" : "SIG OF JSON_ENCODED_DATA"}
        """
        self._fail()
        return None

    def verify_data(self, data):
        """Verify the passed data has been signed by this service. The
           passed data should have the same format as that produced
           by 'sign_data'. If the data is verified then this will
           return a json-deserialised dictionary of the verified data.
           Note that the 'service_uid' should match the UID of this
           service. The data should also contain the fingerprint of the
           key used to encrypt the data, enabling the service to
           perform key rotation and management.
        """
        self._fail()
        return None

    def encrypt_data(self, data):
        """Encrypt the passed data, ready for transport to the service.
           Data should be a json-serialisable dictionary. This will
           return a new json-serialisable dictionary, which will contain
           the UID of the service this should be sent to (together with
           the canonical URL, which enables this data to be forwarded
           to where it needs to go), and the encrypted
           data, e.g. as;

           data = {"service_uid" : "SERVICE_UID",
                   "canonical_url" : "CANONICAL_URL",
                   "fingerprint" : "KEY_FINGERPRINT",
                   "encrypted_data" : "ENCRYPTED_DATA"}
        """
        self._fail()
        return None

    def decrypt_data(self, data):
        """Decrypt the passed data that has been encrypted and sent to
           this service (encrypted via the 'encrypt_data' function).
           This will return a json-deserialisable dictionary. Note that
           the 'service_uid' should match the UID of this
           service. The data should also contain the fingerprint of the
           key used to encrypt the data, enabling the service to
           perform key rotation and management.
        """
        self._fail()
        return None

    def dump_keys(self):
        """Return a dump of the current key and certificate, so that
           we can keep a record of all keys that have been used. The
           returned json-serialisable dictionary contains the keys,
           their fingerprints, and the datetime when they were
           generated. If this is run on the service, then the keys
           are encrypted the password which is encrypted using the
           master key
        """
        self._fail()
        return None

    def load_keys(self, data):
        """Return the keys that were dumped by 'self.dump_keys()'.
           This returns a dictionary of the keys and datetime that
           they were created, indexed by their key fingerprints
        """
        self._fail()
        return None

    def whois(self, username=None, user_uid=None, session_uid=None):
        """Do a whois lookup to map from username to user_uid or
           vice versa. If 'session_uid' is provided, then also validate
           that this is a correct login session, and return also
           the public key and signing certificate for this login session.

           This should return a dictionary with the following keys
           optionally contained;

           username = name of the user
           user_uid = uid of the user
           public_key = public key for the session with uid 'session_uid'
           public_cert = public certificate for that login session
        """
        self._fail()
        return None

    def to_data(self, password=None):
        """Serialise this key to a dictionary, using the supplied
           password to encrypt the private key and certificate"""
        self._fail()
        return None

    @staticmethod
    def from_data(data, password=None):
        """Deserialise this object from the passed data. This will
           only deserialise the private key and private certificate
           if the password is supplied
        """
        from Acquire.Service import Service as _Service
        return _Service.from_data(data, password)




import base64 as _base64
import uuid as _uuid

__all__ = ["UserAccount"]


class UserAccount:
    """This class holds all information about a user's account,
       e.g. their username, the sanitised username for the person
       on the system, their account keys, status etc.

       This data can be serialised to an from json to allow
       easy saving a retrieval from an object store
    """

    def __init__(self, username=None):
        """Construct from the passed username"""
        self._username = username
        self._sanitised_username = UserAccount.sanitise_username(username)
        self._privkey = None
        self._pubkey = None
        self._otp_secret = None
        self._uuid = None

        if username is None:
            self._status = None
        else:
            self._status = "disabled"

    def __str__(self):
        return "UserAccount( name : %s )" % self._username

    def name(self):
        """Return the name of this account"""
        return self._username

    def username(self):
        """Synonym for 'name'"""
        return self.name()

    def sanitised_name(self):
        """Return the sanitised username"""
        return self._sanitised_username

    def uuid(self):
        """Return the globally unique ID for this account"""
        return self._uuid

    def uid(self):
        """Return the globally unique ID for this account"""
        return self.uuid()

    def max_open_sessions(self):
        """Return the maximum number of open login sessions
           (and open login requests) allowed for this user account"""
        return 10

    def login_request_timeout(self):
        """Return the number of hours a login request will
           remain active. This should normally be short, e.g. 30 minutes"""
        return 0.5

    def login_timeout(self):
        """Return the maximum number of hours a single login
           can remain active. This should normally be of the order
           of 1-7 days, as individual calculations or workflows
           should not normally take longer than this"""
        return 7 * 24.0

    def login_root_url(self):
        """Return the root URL used to log into this account"""
        from Acquire.Service import get_this_service as _get_this_service
        return _get_this_service().canonical_url()

    def is_valid(self):
        """Return whether or not this is a valid account"""
        return not (self._status is None)

    def is_active(self):
        """Return whether or not this is an active account"""
        if self._status is None:
            return False
        else:
            return self._status == "active"

    def public_key(self):
        """Return the lines of the public key for this account"""
        return self._pubkey

    def private_key(self):
        """Return the lines of the private key for this account"""
        return self._privkey

    def status(self):
        """Return the status for this account"""
        if self._status is None:
            return "invalid"

        return self._status

    def otp_secret(self):
        """Return the encrypted one-time-password secret"""
        return self._otp_secret

    def set_keys(self, privkey, pubkey, secret=None):
        """Set the private and public keys for this account. The
           keys can be set from files or from a binary read file..
        """
        if self._status is None or privkey is None or pubkey is None:
            return

        try:
            privkey = open(privkey, "rb").read()
        except:
            pass

        try:
            pubkey = open(pubkey, "rb").read()
        except:
            pass

        self._privkey = privkey
        self._pubkey = pubkey
        self._otp_secret = secret

        if self._uuid is None:
            # generate the uuid now, as this should not happen until
            # the account has been first activated. After this point,
            # the uuid of the account should not change
            self._uuid = str(_uuid.uuid4())

        self._status = "active"

    def reset_password(self, password):
        """Call this function to reset the password of this account.
           Note that this will reset the password, returning the
           new OTP used to generate the 2FA one-time-codes
        """
        from Acquire.Crypto import PrivateKey as _PrivateKey
        from Acquire.Crypto import OTP as _OTP

        privkey = _PrivateKey()
        pubkey = privkey.public_key()
        otp = _OTP()

        self.set_keys(privkey.bytes(password), pubkey.bytes(),
                      otp.encrypt(pubkey))

        return otp

    @staticmethod
    def sanitise_username(username):
        """This function returns a sanitised version of
           the username. This will ensure that the username
           is valid (must be between 3 and 50 characters) and
           will remove anything problematic for the object
           store
        """

        if username is None:
            return None

        if len(username) < 3 or len(username) > 50:
            from Acquire.Identity import UsernameError
            raise UsernameError(
                "The username must be between 3 and 50 characters!")

        return "_".join(username.split()).replace("/", "") \
                  .replace("@", "_AT_").replace(".", "_DOT_")

    def validate_password(self, password, otpcode, remember_device=False,
                          device_secret=None):
        """Validate that the passed password and one-time-code are valid.
           If they are, then do nothing. Otherwise raise an exception.
           If 'remember_device' is true, then this returns the provisioning
           uri needed to initialise the OTP code for this account
        """
        if not self.is_active():
            from Acquire.Identity import UserValidationError
            raise UserValidationError(
                "Cannot validate against an inactive account")

        from Acquire.Crypto import PrivateKey as _PrivateKey
        from Acquire.Crypto import OTP as _OTP

        # see if we can decrypt the private key using the password
        privkey = _PrivateKey.read_bytes(self._privkey, password)

        if device_secret:
            # decrypt the passed device secret and check the supplied
            # otpcode for that...
            from Acquire.ObjectStore import string_to_bytes as _string_to_bytes
            otp = _OTP.decrypt(_string_to_bytes(device_secret), privkey)
            otp.verify(otpcode)
        else:
            # now decrypt the secret otp and validate the supplied otpcode
            otp = _OTP.decrypt(self._otp_secret, privkey)
            otp.verify(otpcode)

        if remember_device:
            # create a new OTP that is unique for this device and return
            # this together with the provisioning code
            from Acquire.ObjectStore import bytes_to_string as _bytes_to_string
            otp = _OTP()
            otpsecret = _bytes_to_string(otp.encrypt(privkey.public_key()))
            return (otpsecret, otp.provisioning_uri(self.username()))

    def to_data(self):
        """Return a data representation of this object (dictionary)"""
        if self._username is None:
            return None

        data = {}
        data["username"] = self._username
        data["status"] = self._status
        data["uuid"] = self._uuid

        # the keys and secret are arbitrary binary data.
        # These need to be base64 encoded and then turned into strings
        from Acquire.ObjectStore import bytes_to_string as _bytes_to_string
        data["private_key"] = _bytes_to_string(self._privkey)
        data["public_key"] = _bytes_to_string(self._pubkey)
        data["otp_secret"] = _bytes_to_string(self._otp_secret)

        return data

    @staticmethod
    def from_data(data):
        """Return a UserAccount constructed from the passed
           data (dictionary)
        """

        if data is None:
            return None

        from Acquire.ObjectStore import string_to_bytes as _string_to_bytes

        user_account = UserAccount(data["username"])
        user_account._privkey = _string_to_bytes(data["private_key"])
        user_account._pubkey = _string_to_bytes(data["public_key"])
        user_account._status = data["status"]
        user_account._uuid = data["uuid"]

        try:
            user_account._otp_secret = _string_to_bytes(data["otp_secret"])
        except:
            user_account._otp_secret = None

        return user_account




__all__ = ["create_decimal", "get_decimal_context"]


def get_decimal_context():
    """Return the context used for all decimals in transactions. This
       context rounds to 6 decimal places and provides sufficient precision
       to support any value between 0 and 999,999,999,999,999.999,999,999
       (i.e. everything up to just under one quadrillion - I doubt we will
        ever have an account that has more than a trillion units in it!)
    """
    from decimal import Context as _Context
    return _Context(prec=24)


def create_decimal(value):
    """Create a decimal from the passed value. This is a decimal that
       has 6 decimal places and is clamped between
       -1 quadrillion < value < 1 quadrillion
    """
    from decimal import Decimal as _Decimal

    try:
        d = _Decimal("%.6f" % value, get_decimal_context())
    except:
        value = _Decimal(value, get_decimal_context())
        d = _Decimal("%.6f" % value, get_decimal_context())

    if d <= -1000000000000:
        from Acquire.Accounting import AccountError
        raise AccountError(
                "You cannot create a balance with a value less than "
                "-1 quadrillion! (%s)" % (value))

    elif d >= 1000000000000000:
        from Acquire.Accounting import AccountError
        raise AccountError(
                "You cannot create a balance with a value greater than "
                "1 quadrillion! (%s)" % (value))

    return d




__all__ = ["Receipt"]


class Receipt:
    """This class holds the receipt for a provisional transaction. This is sent
       by the credited account to receipt that the service has been performed,
       and thus payment that is held as liability should now be paid. OR it
       sends back the fact that the service was not performed, and so the
       refund should be issued
    """
    def __init__(self, credit_note=None, authorisation=None,
                 receipted_value=None):
        """Create a receipt for the transaction that resulted in the passed
           credit note. Specify the authorisation of the receipt, and
           optionally specify the actual receipted value (which may be
           less than the value in the passed credit note)
        """
        if credit_note is None:
            self._credit_note = None
            self._authorisation = None
            from Acquire.Accounting import create_decimal as _create_decimal
            self._receipted_value = _create_decimal(0)
            return

        from Acquire.Accounting import CreditNote as _CreditNote
        from Acquire.Identity import Authorisation as _Authorisation

        if not isinstance(credit_note, _CreditNote):
            raise TypeError("The credit note must be of type CreditNote")

        if not isinstance(authorisation, _Authorisation):
            raise TypeError("The authorisation must be of type Authorisation")

        if not credit_note.is_provisional():
            raise ValueError("You cannot receipt a transaction that was "
                             "not provisional! - %s" % str(credit_note))

        if receipted_value is not None:
            from Acquire.Accounting import create_decimal as _create_decimal
            receipted_value = _create_decimal(receipted_value)

            if receipted_value < 0:
                raise ValueError("You cannot receipt a value that is less "
                                 "than zero! %s" % receipted_value)
            elif receipted_value > credit_note.value():
                raise ValueError("You cannot receipt a value that is greater "
                                 "than the value of the original credit "
                                 "note - %s versus %s" % (receipted_value,
                                                          credit_note.value()))
        else:
            receipted_value = credit_note.value()

        self._credit_note = credit_note
        self._authorisation = authorisation
        self._receipted_value = receipted_value

    def __str__(self):
        return "Receipt(credit_note=%s, receipted_value=%s)" % \
            (str(self.credit_note()), self.receipted_value())

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self._credit_note == other._credit_note and \
                   self._receipted_value == other._receipted_value and \
                   self._authorisation == other._authorisation
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def is_null(self):
        """Return whether or not this Receipt is null"""
        return self._credit_note is None

    def credit_note(self):
        """Return the credit note that this is a receipt for"""
        if self.is_null():
            from Acquire.Accounting import CreditNote as _CreditNote
            return _CreditNote()
        else:
            return self._credit_note

    def transaction_uid(self):
        """Return the UID of the provisional transaction for which this
           is the receipt. The transaction UID is the same as the UID
           for the original debit note
        """
        return self.debit_note_uid()

    def debit_note_uid(self):
        """Return the UID of the debit note that this is a receipt for"""
        if self.is_null():
            return None
        else:
            return self._credit_note.debit_note_uid()

    def debit_account_uid(self):
        """Return the UID of the account from which this receipt
           will debit value
        """
        if self.is_null():
            return None
        else:
            return self._credit_note.debit_account_uid()

    def credit_account_uid(self):
        """Return the UID of the account to which this receipt will
           credit value
        """
        if self.is_null():
            return None
        else:
            return self._credit_note.credit_account_uid()

    def transaction(self):
        """Return a transaction that corresponds to the real transfer
           of value between the debit and credit accounts. The value
           of the transaction is the actual receipted value
        """
        from Acquire.Accounting import Transaction as _Transaction
        if self.is_null():
            return _Transaction()
        else:
            return _Transaction(self.receipted_value(),
                                "Receipt for transaction %s"
                                % self.transaction_uid())

    def value(self):
        """Return the original (provisional) value of the transaction"""
        if self.is_null():
            from Acquire.Accounting import create_decimal as _create_decimal
            return _create_decimal(0)
        else:
            return self._credit_note.value()

    def provisional_value(self):
        """Return the original (provisional) value of the transaction"""
        return self.value()

    def receipted_value(self):
        """Return the receipted value. This is guaranteed to be less than
           or equal to the provisional value in the attached CreditNote
        """
        if self.is_null():
            from Acquire.Accounting import create_decimal as _create_decimal
            return _create_decimal(0)
        else:
            return self._receipted_value

    def authorisation(self):
        """Return the authorisation for the receipt"""
        return self._authorisation

    def to_data(self):
        """Return the data for this object as a dictionary that can be
           serialised to json
        """
        data = {}

        if not self.is_null():
            data["credit_note"] = self._credit_note.to_data()
            data["authorisation"] = self._authorisation.to_data()
            data["receipted_value"] = str(self._receipted_value)

        return data

    @staticmethod
    def from_data(data):
        """Return a Receipt from the passed json-decoded dictionary"""
        r = Receipt()

        if (data and len(data) > 0):
            from Acquire.Accounting import CreditNote as _CreditNote
            from Acquire.Accounting import create_decimal as _create_decimal
            from Acquire.Identity import Authorisation as _Authorisation
            r._credit_note = _CreditNote.from_data(data["credit_note"])
            r._authorisation = _Authorisation.from_data(data["authorisation"])
            r._receipted_value = _create_decimal(data["receipted_value"])

        return r

    @staticmethod
    def create(credit_notes, authorisation, receipted_value=None):
        """Construct a series of receipts from the passed credit notes,
           each of which is authorised using the passed authorisation.
           If 'receipted_value' is specified, then the sum total of
           receipts will equal 'receipted_value'. This cannot be
           greater than the sum of the passed credit notes. If it
           is less then the value, then the difference is subtracted
           from the first receipts returned
        """
        try:
            credit_note = credit_notes[0]
        except:
            return Receipt(credit_notes, authorisation, receipted_value)

        if len(credit_notes) == 0:
            return Receipt()
        elif len(credit_notes) == 1:
            return Receipt(credit_notes[0], authorisation, receipted_value)

        from Acquire.Accounting import create_decimal as _create_decimal
        total_value = _create_decimal(0)

        from Acquire.Accounting import CreditNote as _CreditNote

        for credit_note in credit_notes:
            if not isinstance(credit_note, _CreditNote):
                raise TypeError("The credit note must be of type CreditNote")

            total_value += credit_note.value()

        if receipted_value is None:
            receipted_value = total_value
        else:
            receipted_value = _create_decimal(receipted_value)

        if receipted_value < 0:
            raise ValueError("You cannot receipt a value that is less "
                             "than zero! %s" % receipted_value)
        elif receipted_value > total_value:
            raise ValueError("You cannot receipt a value that is greater "
                             "than the value of the original credit "
                             "notes - %s versus %s" % (receipted_value,
                                                       total_value))

        receipts = []

        if receipted_value == total_value:
            for credit_note in credit_notes:
                receipts.append(Receipt(credit_note, authorisation))
        else:
            diff = total_value - receipted_value

            for credit_note in credit_notes:
                d = min(diff, credit_note.value())

                diff -= d
                r = credit_note.value() - d

                receipts.append(Receipt(credit_note, authorisation, r))

        return receipts



"""
Acquire: (C) Christopher Woods 2018

This module implements all of the classes and functions necessary to build the
client (user-facing) interfaces for Acquire
"""

from ._qrcode import *
from ._user import *
from ._account import *
from ._drive import *
from ._filehandle import *
from ._fileops import *
from ._wallet import *
from ._errors import *
from ._cheque import *
from ._service import *
from ._par import *

# The below objects are useful for the client, so are pulled into
# this module to discourage people using the other Acquire modules
# directly... One day I want to lazy-load all of these...
from Acquire.Crypto import PublicKey
from Acquire.Crypto import PrivateKey
from Acquire.Crypto import OTP
from Acquire.Identity import Authorisation
from Acquire.Storage import ACLRule
from Acquire.Storage import FileMeta
from Acquire.Storage import DriveMeta

try:
    if __IPYTHON__:
        def _set_printer(C):
            """Function to tell ipython to use __str__ if available"""
            get_ipython().display_formatter.formatters['text/plain'].for_type(
                C,
                lambda obj, p, cycle: p.text(str(obj) if not cycle else '...')
                )

        import sys as _sys
        import inspect as _inspect

        _clsmembers = _inspect.getmembers(_sys.modules[__name__],
                                          _inspect.isclass)

        for _clsmember in _clsmembers:
            _set_printer(_clsmember[1])
except:
    pass




__all__ = ["PairedNote"]


class PairedNote:
    """This class holds a DebitNote together with its matching
       CreditNote(s)
    """
    def __init__(self, debit_note, credit_note):
        """Construct from the matching pair of notes"""
        from Acquire.Accounting import CreditNote as _CreditNote
        from Acquire.Accounting import DebitNote as _DebitNote

        if not isinstance(debit_note, _DebitNote):
            raise TypeError("The debit_note must be of type DebitNote!")

        if not isinstance(credit_note, _CreditNote):
            raise TypeError("The credit_note must be of type CreditNote!")

        if credit_note.debit_note_uid() != debit_note.uid():
            raise ValueError("You must pair up DebitNote (%s) with a "
                             "matching CreditNote (%s)" %
                             (debit_note.uid(), credit_note.debit_note_uid()))

        self._debit_note = debit_note
        self._credit_note = credit_note

    def __str__(self):
        return "PairedNote(debit_note=%s, credit_note=%s)" % \
                    (str(self._debit_note), str(self._credit_note))

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self._debit_note == other._debit_note and \
                   self._credit_note == other._credit_note
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def debit_note(self):
        """Return the debit note"""
        return self._debit_note

    def credit_note(self):
        """Return the credit note"""
        return self._credit_note

    @staticmethod
    def create(debit_notes, credit_notes):
        """Return a list of PairedNotes that pair together the passed
           debit notes and credit notes
        """
        try:
            debit_note = debit_notes[0]
        except:
            debit_notes = [debit_notes]

        if not isinstance(credit_notes, dict):
            try:
                credit_notes[0]
            except:
                credit_notes = [credit_notes]

            d = {}
            for credit_note in credit_notes:
                d[credit_note.debit_note_uid()] = credit_note
            credit_notes = d

        pairs = []
        missing = []

        for debit_note in debit_notes:
            if debit_note.uid() in credit_notes:
                pairs.append(PairedNote(debit_note,
                             credit_notes[debit_note.uid()]))
            else:
                missing.append(debit_note)

        if len(missing) > 0 or len(credit_notes) != len(debit_notes):
            from Acquire.Accounting import UnbalancedLedgerError
            raise UnbalancedLedgerError(
                "Cannot balance the ledger as the debit do not match the "
                "credits %s versus %s" % (str(debit_notes), str(credit_notes)))

        return pairs




import datetime as _datetime
import json as _json

__all__ = ["Account", "get_accounts", "create_account",
           "deposit", "withdraw"]


def _get_accounting_url():
    """Function to discover and return the default accounting url"""
    return "http://fn.acquire-aaai.com:8080/t/accounting"


def _get_accounting_service(accounting_url=None):
    """Function to return the accounting service for the system"""
    if accounting_url is None:
        accounting_url = _get_accounting_url()

    from Acquire.Client import Service as _Service
    service = _Service(accounting_url)

    if not service.is_accounting_service():
        from Acquire.Client import LoginError
        raise LoginError(
            "You can only use a valid accounting service to get account info! "
            "The service at '%s' is a '%s'" %
            (accounting_url, service.service_type()))

    if service.service_url() != accounting_url:
        service.update_service_url(accounting_url)

    return service


def _get_account_uid(user, account_name, accounting_service=None,
                     accounting_url=None):
    """Return the UID of the account called 'account_name' that
        belongs to passed user on the passed accounting_service
    """
    if account_name is None:
        # return the UID of the default account for this user
        account_name = "main"

    if accounting_service is None:
        service = _get_accounting_service(accounting_url)
    else:
        if not accounting_service.is_accounting_service():
            raise TypeError("You can only query accounts using "
                            "a valid accounting service")
        service = accounting_service

    args = {"user_uid": user.uid(),
            "account_name": str(account_name)}

    if user.is_logged_in():
        from Acquire.Client import Authorisation as _Authorisation
        auth = _Authorisation(user=user)
        args["authorisation"] = auth.to_data()

    result = service.call_function(function="get_account_uids", args=args)

    account_uids = result["account_uids"]

    for account_uid in account_uids:
        if account_uids[account_uid] == account_name:
            return account_uid

    from Acquire.Client import AccountError
    raise AccountError("There is no account called '%s' for '%s'" %
                       (account_name, str(user)))


def _get_account_uids(user, accounting_service=None, accounting_url=None):
    """Return the names and UIDs of all of the accounts that belong
        to the passed user on the passed accounting_service
    """
    if accounting_service is None:
        service = _get_accounting_service(accounting_url)
    else:
        if not accounting_service.is_accounting_service():
            raise TypeError("You can only query accounts using "
                            "a valid accounting service")
        service = accounting_service

    if not user.is_logged_in():
        raise PermissionError(
            "You can only get information about about a user's accounts "
            "if they have authenticated their login")

    from Acquire.Client import Authorisation as _Authorisation
    auth = _Authorisation(user=user)
    args = {"authorisation": auth.to_data()}

    result = service.call_function(function="get_account_uids", args=args)

    return result["account_uids"]


def get_accounts(user, accounting_service=None, accounting_url=None):
    """Return all of the accounts of the passed user. Note that the
    user must be authenticated to call this function
    """
    if accounting_service is None:
        service = _get_accounting_service(accounting_url)
    else:
        if not accounting_service.is_accounting_service():
            raise TypeError("You can only query account using "
                            "a valid accounting service")
        service = accounting_service

    account_uids = _get_account_uids(
                        user, accounting_service=service)

    accounts = []

    for uid in account_uids.keys():
        name = account_uids[uid]

        account = Account()
        account._account_name = name
        account._account_uid = uid
        account._user = user
        account._accounting_service = service

        accounts.append(account)

    return accounts


def create_account(user, account_name, description=None,
                   accounting_service=None, accounting_url=None):
    """Create an account on the accounting service for the passed
        user, calling the account 'account_name' and optionally
        passing in an account description. Note that the user must
        have authorised the login
    """
    if accounting_service is None:
        service = _get_accounting_service(accounting_url)
    else:
        if not accounting_service.is_accounting_service():
            raise TypeError("You can only query account using "
                            "a valid accounting service")
        service = accounting_service

    if not user.is_logged_in():
        raise PermissionError(
            "You cannot create an account called '%s' for user "
            "'%s' as the user login has not been authenticated." %
            (account_name, user.name()))

    from Acquire.Client import Authorisation as _Authorisation
    authorisation = _Authorisation(user=user)

    args = {"account_name": str(account_name),
            "authorisation": authorisation.to_data()}

    if description is None:
        args["description"] = "Account '%s' for '%s'" % \
                                (str(account_name), user.name())
    else:
        args["description"] = str(description)

    result = service.call_function(function="create_account", args=args)

    account_uid = result["account_uid"]

    account = Account()
    account._account_name = account_name
    account._account_uid = account_uid
    account._user = user
    account._accounting_service = service

    return account


def deposit(user, value, description=None,
            accounting_service=None, accounting_url=None):
    """Tell the system to allow the user to deposit 'value' from
       their (real) financial account to the system accounts
    """
    from Acquire.Client import Authorisation as _Authorisation
    authorisation = _Authorisation(user=user)

    if accounting_service is None:
        service = _get_accounting_service(accounting_url)
    else:
        if not accounting_service.is_accounting_service():
            raise TypeError("You can only deposit funds using an "
                            "accounting service!")
        service = accounting_service

    args = {"authorisation": authorisation.to_data()}

    if description is None:
        from Acquire.Accounting import create_decimal as _create_decimal
        args["value"] = str(_create_decimal(value))
    else:
        from Acquire.Accounting import Transaction as _Transaction
        args["transaction"] = _Transaction(value, description).to_data()

    result = service.call_function(function="deposit", args=args)

    return result


def withdraw(user, value, description=None,
             accounting_service=None, accounting_url=None):
    """Tell the system to allow the user to withdraw 'value' from
       the system accounts back to their (real) financial account
    """
    raise NotImplementedError("withdrawals are not yet implemented!")


class Account:
    """This is the client-side handle that is used to interact with
       an account on the service. If the account is created with a valid
       user login then you can perform tasks such as making payments,
       or issueing receipts or refunds. Otherwise, this is a simple
       interface that allows the account to be used as a receiver
       of value
    """
    def __init__(self, user=None, account_name=None, accounting_service=None,
                 accounting_url=None):
        """Construct the Account with the passed account_name, which is owned
           by the passed user. The account must already exist on the service,
           or else an exception will be raised
        """
        if user is not None:
            if account_name is None:
                self._account_name = "main"
            else:
                self._account_name = str(account_name)

            self._user = user

            if accounting_service is None:
                accounting_service = _get_accounting_service(accounting_url)
            else:
                from Acquire.Accounting import AccountingService \
                    as _AccountingService

                if not isinstance(accounting_service, _AccountingService):
                    raise TypeError(
                        "You can only create an account on a valid "
                        "AccountingService object")

                if not accounting_service.is_accounting_service():
                    raise TypeError(
                        "The passed service - %s - is not an accounting "
                        "service." % str(accounting_service))

            self._accounting_service = accounting_service

            self._account_uid = _get_account_uid(user, account_name,
                                                 accounting_service)
        else:
            self._account_uid = None

        self._last_update = None
        self._description = None

    def __str__(self):
        if self.is_null():
            return "Account::null"
        else:
            return "Account(name='%s', uid=%s)" % (self.name(), self.uid())

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self._account_uid == other._account_uid
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def is_null(self):
        """Return whether or not this is a null account"""
        return self._account_uid is None

    def uid(self):
        """Return the UID of this account"""
        return self._account_uid

    def guid(self):
        """Return the globally unique UID of this account. This is a
           combination of the UID of the accounting service and the
           UID of the account
        """
        return "%s@%s" % (self.accounting_service().uid(), self.uid())

    def name(self):
        """Return the name of this account"""
        if self.is_null():
            return None
        else:
            return self._account_name

    def owner(self):
        """Return the user who owns this account"""
        if self.is_null():
            return None
        else:
            return self._user

    def user(self):
        """Synonym for owner"""
        return self.owner()

    def is_logged_in(self):
        """Return whether or not the user has an authenticated login
           to this account
        """
        try:
            return self._user.is_logged_in()
        except:
            return False

    def last_update_time(self):
        """Return the time of the last update of the balance"""
        return self._last_update

    def _refresh(self, force_update=False):
        """Refresh the current status of this account. This fetches
           the latest data, e.g. balance, limits etc. Note that this
           limits you to refreshing at most once every five seconds...
        """
        if self.is_null():
            from Acquire.Accounting import create_decimal as _create_decimal
            self._overdraft_limit = _create_decimal(0)
            self._balance = _create_decimal(0)
            self._liability = _create_decimal(0)
            self._receivable = _create_decimal(0)
            self._spent_today = _create_decimal(0)
            return

        if force_update:
            should_refresh = True
        else:
            should_refresh = False

            if self._last_update is None:
                should_refresh = True
            else:
                should_refresh = (_datetime.datetime.now() -
                                  self._last_update).seconds > 5

        if not should_refresh:
            return

        if not self.is_logged_in():
            raise PermissionError(
                "You cannot get information about this account "
                "until after the owner has successfully authenticated.")

        from Acquire.Client import Authorisation as _Authorisation
        from Acquire.Accounting import create_decimal as _create_decimal

        service = self.accounting_service()

        auth = _Authorisation(resource=self._account_uid, user=self._user)

        args = {"authorisation": auth.to_data(),
                "account_name": self.name()}

        result = service.call_function(function="get_info", args=args)

        self._overdraft_limit = _create_decimal(result["overdraft_limit"])
        self._balance = _create_decimal(result["balance"])
        self._liability = _create_decimal(result["liability"])
        self._receivable = _create_decimal(result["receivable"])
        self._spent_today = _create_decimal(result["spent_today"])
        self._description = result["description"]

        self._last_update = _datetime.datetime.now()

    def accounting_service(self):
        """Return the accounting service managing this account"""
        return self._accounting_service

    def description(self):
        """Return the description of this account"""
        if not self._description:
            self._refresh()

        return self._description

    def balance(self, force_update=False):
        """Return the current balance of this account"""
        self._refresh(force_update)
        return self._balance

    def liability(self, force_update=False):
        """Return the current total liability of this account"""
        self._refresh(force_update)
        return self._liability

    def receivable(self, force_update=False):
        """Return the current total accounts receivable of this account"""
        self._refresh(force_update)
        return self._receivable

    def spent_today(self, force_update=False):
        """Return the current amount spent today on this account"""
        self._refresh(force_update)
        return self._spent_today

    def overdraft_limit(self, force_update=False):
        """Return the overdraft limit of this account"""
        self._refresh(force_update)
        return self._overdraft_limit

    def is_beyond_overdraft_limit(self, force_update=False):
        """Return whether or not the current balance is beyond
           the overdraft limit
        """
        self._refresh(force_update)
        return (self._balance - self._liability) < -(self._overdraft_limit)

    def perform(self, transaction, credit_account, is_provisional=False):
        """Tell this accounting service to apply the transfer described
           in 'transaction' from this account to the passed account. Note
           that the user must have logged into this account so that they
           have authorised this transaction. This returns the record
           of this transaction
        """
        if not self.is_logged_in():
            raise PermissionError("You cannot transfer value from '%s' to "
                                  "'%s' because you have not authenticated "
                                  "the user who owns this account" %
                                  (str(self), str(credit_account)))

        from Acquire.Accounting import Transaction as _Transaction

        if not isinstance(transaction, _Transaction):
            raise TypeError("The passed transaction must be of type "
                            "Transaction")

        if not isinstance(credit_account, Account):
            raise TypeError("The passed credit account must be of type "
                            "Account")

        if transaction.is_null():
            return None

        from Acquire.Client import Authorisation as _Authorisation
        service = self.accounting_service()

        auth = _Authorisation(resource=self._account_uid, user=self._user)

        if is_provisional:
            is_provisional = True
        else:
            is_provisional = False

        args = {"transaction": transaction.to_data(),
                "debit_account_uid": str(self.uid()),
                "credit_account_uid": str(credit_account.uid()),
                "is_provisional": is_provisional,
                "authorisation": auth.to_data()}

        result = service.call_function(function="perform", args=args)

        return result["transaction_records"]

    def receipt(self, credit_note, receipted_value=None):
        """Receipt the passed credit note that contains a request to
           transfer value from another account to the passed account
        """
        if not self.is_logged_in():
            raise PermissionError("You cannot receipt a credit note as the "
                                  "user has not yet logged in!")

        if credit_note.account_uid() != self.uid():
            raise ValueError(
                "You cannot receipt a transaction from a different "
                "account! %s versus %s" % (credit_note.account_uid(),
                                           self.uid()))

        from Acquire.Client import Authorisation as _Authorisation
        from Acquire.Accounting import create_decimal as _create_decimal

        service = self.accounting_service()

        auth = _Authorisation(resource=self._account_uid, user=self._user)

        args = {"credit_note": credit_note.to_data(),
                "authorisation": auth.to_data()}

        if receipted_value is not None:
            args["receipted_value"] = str(_create_decimal(receipted_value))

        result = service.call_function(function="receipt", args=args)

        return result["transaction_record"]

    def refund(self, credit_note):
        """Refunds the passed credit note that contained a transfer of
           from another account to the passed account
        """
        if not self.is_logged_in():
            raise PermissionError("You cannot refund a credit note as the "
                                  "user has not yet logged in!")

        if credit_note.account_uid() != self.uid():
            raise ValueError(
                "You cannot refund a transaction from a different "
                "account! %s versus %s" % (credit_note.account_uid(),
                                           self.uid()))

        from Acquire.Client import Authorisation as _Authorisation
        from Acquire.Accounting import create_decimal as _create_decimal

        service = self.accounting_service()

        auth = _Authorisation(resource=self._account_uid, user=self._user)

        args = {"credit_note": credit_note.to_data(),
                "authorisation": auth.to_data()}

        result = service.call_function(function="refund", args=args)

        return result["transaction_record"]





__all__ = ["ACLRule"]


def _max(a, b):
    """Function to do most-permissive addition of two bools

        True + True = True
        True + False = True
        False + True = True
        False + False = False

        True + None = True
        None + True = True

        False + None = False
        None + False = False

        None + None = None
    """
    if a is None:
        return b
    elif b is None:
        return a
    elif a is True:
        return True
    else:
        return b


def _sub(a, b):
    """Function to do subtraction of two bools

        True - True = False
        True - False = True
        False - True = False
        False - False = False

        True - None = True
        None - True = False

        False - None = False
        None - False = None

        None - None = None
    """
    if b is True:
        return False
    else:
        return a


def _min(a, b):
    """Function to do least-permission addition of two bools

        True * True = True
        True * False = False
        False * True = False
        False * False = False

        True * None = True
        None * True = True

        False * None = False
        None * False = False

        None - None = None
    """
    if a is None:
        return b
    elif b is None:
        return a
    elif a is False:
        return False
    else:
        return b


class ACLRule:
    """This class holds the access control list (ACL) rule for
       a particular user accessing a particular resource
    """
    def __init__(self, is_owner=None, is_readable=None,
                 is_writeable=None, is_executable=None):
        """Construct a default rule. By default this rule has
           fully-inherit permissions (nothing is set either way)
        """
        if is_owner is None:
            self._is_owner = None
        elif is_owner:
            self._is_owner = True
        else:
            self._is_owner = False

        if is_readable is None:
            self._is_readable = None
        elif is_readable:
            self._is_readable = True
        else:
            self._is_readable = False

        if is_writeable is None:
            self._is_writeable = None
        elif is_writeable:
            self._is_writeable = True
        else:
            self._is_writeable = False

        if is_executable is None:
            self._is_executable = None
        elif is_executable:
            self._is_executable = True
        else:
            self._is_executable = False

    def __str__(self):
        s = []

        if self.is_owner():
            s.append("owner")
        elif self.inherits_owner():
            s.append("inherits_owner")

        if self.is_writeable():
            s.append("writeable")
        elif self.inherits_writeable():
            s.append("inherits_writeable")

        if self.is_readable():
            s.append("readable")
        elif self.inherits_readable():
            s.append("inherits_readable")

        if self.is_executable():
            s.append("executable")
        elif self.inherits_executable():
            s.append("inherits_executable")

        if len(s) == 0:
            return "ACLRule::denied"
        else:
            return "ACLRule(%s)" % ", ".join(s)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        else:
            return False

    def __add__(self, other):
        """Add two rules together - this will combine the parts
           additively, i.e. the most-permissive options are set
           in the return value
        """
        if not isinstance(other, ACLRule):
            raise TypeError(
                "You can only combine together pairs of ACLRules")

        return ACLRule(
                is_owner=_max(self._is_owner, other._is_owner),
                is_writeable=_max(self._is_writeable, other._is_writeable),
                is_readable=_max(self._is_readable, other._is_readable),
                is_executable=_max(self._is_executable, other._is_executable))

    def __sub__(self, other):
        """Subtract 'other' from these rules - if 'other' is False
           the this is set to False
        """
        if not isinstance(other, ACLRule):
            raise TypeError(
                "You can only combine together pairs of ACLRules")

        return ACLRule(
                is_owner=_sub(self._is_owner, other._is_owner),
                is_writeable=_sub(self._is_writeable, other._is_writeable),
                is_readable=_sub(self._is_readable, other._is_readable),
                is_executable=_sub(self._is_executable, other._is_executable))

    def __mul__(self, other):
        """Combine two rules together - this will combine
           the parts subtractively, i.e. the least permissive options
           are set in the return value
        """
        if not isinstance(other, ACLRule):
            raise TypeError(
                "You can only combine together pairs of ACLRules")

        return ACLRule(
                is_owner=_min(self._is_owner, other._is_owner),
                is_writeable=_min(self._is_writeable, other._is_writeable),
                is_readable=_min(self._is_readable, other._is_readable),
                is_executable=_min(self._is_executable, other._is_executable))

    @staticmethod
    def owner():
        """Return the ACLRule of an owner"""
        return ACLRule(is_owner=True, is_readable=True,
                       is_writeable=True, is_executable=True)

    @staticmethod
    def executer():
        """Return the ACLRule of an executer"""
        return ACLRule(is_owner=False, is_executable=True,
                       is_writeable=True, is_readable=True)

    @staticmethod
    def writer():
        """Return the ACLRule of a writer"""
        return ACLRule(is_owner=False, is_readable=True,
                       is_writeable=True, is_executable=False)

    @staticmethod
    def reader():
        """Return the ACLRule of a reader"""
        return ACLRule(is_owner=False, is_readable=True,
                       is_writeable=False, is_executable=False)

    @staticmethod
    def denied():
        """Return a "denied" (no-access at all) permission rule"""
        return ACLRule(is_owner=False, is_readable=False,
                       is_writeable=False, is_executable=False)

    @staticmethod
    def null():
        """Return a null (inherit all permissions) rule"""
        return ACLRule(is_owner=None, is_readable=None,
                       is_writeable=None, is_executable=None)

    @staticmethod
    def inherit():
        """Return the ACL rule that means 'inherit permissions from parent'"""
        return ACLRule(is_owner=None, is_readable=None,
                       is_writeable=None, is_executable=None)

    def is_owner(self):
        """Return whether or not the user is the owner of this resource"""
        return self._is_owner

    def is_readable(self):
        """Return whether or not the user can read this resource"""
        return self._is_readable

    def is_writeable(self):
        """Return whether or not the user can write to this resource"""
        return self._is_writeable

    def is_executable(self):
        """Return whether or not the user can execute this resource"""
        return self._is_executable

    def inherits_owner(self):
        """Return whether or not this inherits the owner status
           from upstream
        """
        return self._is_owner is None

    def inherits_readable(self):
        """Return whether or not this inherits the reader status
           from upstream
        """
        return self._is_readable is None

    def inherits_writeable(self):
        """Return whether or not this inherits the writeable status
           from upstream
        """
        return self._is_writeable is None

    def inherits_executable(self):
        """Return whether or not this inherits the executable status
           from upstream
        """
        return self._is_executable is None

    def inherits_all(self):
        """Return whether or not this rule inherits all permissions"""
        return (self._is_owner is None) and \
               (self._is_readable is None) and \
               (self._is_writeable is None) and \
               (self._is_executable is None)

    def is_fully_resolved(self):
        """Return whether or not this rule is fully resolved"""
        return (self._is_owner is not None) and \
               (self._is_readable is not None) and \
               (self._is_writeable is not None) and \
               (self._is_executable is not None)

    def denied_all(self):
        """Return whether or not this rule shows that everything is
           denied
        """
        return (self._is_owner is False) and \
               (self._is_readable is False) and \
               (self._is_writeable is False) and \
               (self._is_executable is False)

    def _force_resolve(self, unresolved=False):
        """This returns a fully resolved ACL, where anything that
           is not resolved is set equal to 'unresolved'
        """
        if unresolved:
            unresolved = True
        else:
            unresolved = False

        is_owner = self._is_owner
        is_readable = self._is_readable
        is_writeable = self._is_writeable
        is_executable = self._is_executable

        if is_owner is None:
            is_owner = unresolved

        if is_readable is None:
            is_readable = unresolved

        if is_executable is None:
            is_executable = unresolved

        if is_writeable is None:
            is_writeable = unresolved

        return ACLRule(is_owner=is_owner, is_writeable=is_writeable,
                       is_readable=is_readable, is_executable=is_executable)

    def resolve(self, must_resolve=True, **kwargs):
        """Resolve these rules based on the information supplied
           in 'kwargs'. Notably, if any of our rules are 'inherit',
           the this will look for an ACLRule called "upstream" to
           inherit the rule. If 'must_resolve' is true, then
           this function must always return
           a fully-resolved ACLRule
        """
        if self.is_fully_resolved():
            return self

        if "unresolved" in kwargs:
            unresolved = kwargs["unresolved"]
        else:
            unresolved = False

        if "upstream" not in kwargs:
            if must_resolve:
                return self._force_resolve(unresolved=unresolved)
            else:
                return self

        upstream = kwargs["upstream"]

        if not upstream.is_fully_resolved():
            del kwargs["upstream"]
            upstream = upstream.resolve(must_resolve=must_resolve, **kwargs)

        if must_resolve and (not upstream.is_fully_resolved()):
            upstream = upstream._force_resolve(unresolved=unresolved)

        is_owner = self._is_owner
        is_readable = self._is_readable
        is_writeable = self._is_writeable
        is_executable = self._is_executable

        if is_owner is None:
            is_owner = upstream._is_owner

        if is_readable is None:
            is_readable = upstream._is_readable

        if is_writeable is None:
            is_writeable = upstream._is_writeable

        if is_executable is None:
            is_executable = upstream._is_executable

        return ACLRule(is_owner=is_owner, is_writeable=is_writeable,
                       is_readable=is_readable, is_executable=is_executable)

    def set_owner(self, is_owner=True):
        """Set the user as an owner of the bucket"""
        if is_owner:
            self._is_owner = True
        else:
            self._is_owner = False

    def set_readable(self, is_readable=True):
        """Set the readable rule to 'is_readable'"""
        if is_readable:
            self._is_readable = True
        else:
            self._is_readable = False

    def set_writeable(self, is_writeable=True):
        """Set the writeable rule to 'is_writeable'"""
        if is_writeable:
            self._is_writeable = True
        else:
            self._is_writeable = False

    def set_inherits_owner(self):
        """Set that this ACL inherits ownership from its parent"""
        self._is_owner = None

    def set_inherits_readable(self):
        """Set that this ACL inherits readable from its parent"""
        self._is_readable = None

    def set_inherits_writeable(self):
        """Set that this ACL inherits writeable from its parent"""
        self._is_writeable = None

    def set_readable_writeable(self, is_readable_writeable=True):
        """Set both the readable and writeable rules to
           'is_readable_writeable'
        """
        if is_readable_writeable:
            self._is_readable = True
            self._is_writeable = True
        else:
            self._is_readable = False
            self._is_writeable = False

    def to_data(self):
        """Return this object converted to a json-serialisable object"""
        if self.inherits_all():
            return "inherits"
        elif self == ACLRule.owner():
            return "owner"
        elif self == ACLRule.reader():
            return "reader"
        elif self == ACLRule.writer():
            return "writer"
        elif self == ACLRule.denied():
            return "denied"
        elif self == ACLRule.executer():
            return "executer"
        else:
            data = {}
            data["is_owner"] = self._is_owner
            data["is_readable"] = self._is_readable
            data["is_writeable"] = self._is_writeable
            return data

    @staticmethod
    def from_data(data):
        """Return this object constructed from the passed json-deserialised
           dictionary
        """
        if data is None:
            return None

        if isinstance(data, str):
            if data == "inherits":
                return ACLRule.inherit()
            elif data == "owner":
                return ACLRule.owner()
            elif data == "reader":
                return ACLRule.reader()
            elif data == "writer":
                return ACLRule.writer()
            elif data == "executer":
                return ACLRule.executer()
            elif data == "denied":
                return ACLRule.denied()

        is_owner = None
        is_readable = None
        is_writeable = None

        if "is_owner" in data:
            if data["is_owner"]:
                is_owner = True
            elif data["is_owner"] is not None:
                is_owner = False

        if "is_readable" in data:
            if data["is_readable"]:
                is_readable = True
            elif data["is_readable"] is not None:
                is_readable = False

        if "is_writeable" in data:
            if data["is_writeable"]:
                is_writeable = True
            elif data["is_writeable"] is not None:
                is_writeable = False

        return ACLRule(is_owner=is_owner,
                       is_writeable=is_writeable,
                       is_readable=is_readable)




__all__ = ["OTP"]


class OTP:
    """This class handles everything to do with obtaining and
       verifying a one-time-password"""
    def __init__(self, secret=None):
        """This creates a new one-time-password"""
        try:
            import pyotp as _pyotp
        except:
            from Acquire.Crypto import OTPError
            raise OTPError(
                "Cannot create a one-time-password as the "
                "pyotp module is not available. Please install and try again")

        if secret:
            self._secret = secret
        else:
            self._secret = _pyotp.random_base32()

    def __str__(self):
        """Return a string representation of this OTP"""
        return "OTP()"

    @staticmethod
    def decrypt(secret, key):
        """Construct a OTP from the passed encrypted secret
           that will be decrypted with the passed private key"""
        from Acquire.Crypto import PrivateKey as _PrivateKey
        if not isinstance(key, _PrivateKey):
            raise TypeError(
                "You can only encrypt a OTP using a valid "
                "PrivateKey - not using a %s" % key.__class__)

        otp = OTP()
        otp._secret = key.decrypt(secret)

        return otp

    def encrypt(self, key):
        """This uses the passed public key to encrypt and return the
           secret"""
        from Acquire.Crypto import PublicKey as _PublicKey
        if not isinstance(key, _PublicKey):
            raise TypeError(
                "You can only encrypt a OTP using a valid "
                "PublicKey - not using a %s" % key.__class__)

        return key.encrypt(self._secret)

    def _totp(self):
        """Return the time-based one-time-password based on this secret"""
        try:
            import pyotp as _pyotp
            return _pyotp.totp.TOTP(self._secret)
        except:
            from Acquire.Crypto import OTPError
            raise OTPError("You cannot get a null OTP - create one first!")

    def provisioning_uri(self, username, issuer="Acquire"):
        """Return the provisioning URI, assuming this secret is
           for the user called 'username' and is issued by 'issuer'"""
        return self._totp().provisioning_uri(username, issuer_name=issuer)

    @staticmethod
    def extract_secret(provisioning_uri):
        """Return the otpsecret extracted from the passed provisioning_url"""
        import re as _re
        try:
            return _re.search(r"secret=([\w\d+]+)&issuer",
                              provisioning_uri).groups()[0]
        except Exception as e:
            from Acquire.Crypto import OTPError
            raise OTPError(
                "Cannot extract the otp secret from the provisioning URL "
                "'%s': %s" % (provisioning_uri, str(e)))

    def generate(self):
        """Generate and return the current OTP code"""
        totp = self._totp()
        return totp.now()

    def verify(self, code):
        """Verify that the passed code is correct. This raises an exception
           if the code is incorrect, or does nothing if the code is correct"""

        # the OTP is valid for 1 minute. We will extend this so that
        # it is valid for 3 minutes (1 minute before and after). This
        # improves usability and tolerance for clock drift with only
        # minor increase in OTP validity time
        if not self._totp().verify(code, valid_window=1):
            from Acquire.Crypto import OTPError
            raise OTPError("The passed OTP code is incorrect")

        # note that, ideally, we need to save whether or not this code
        # has been used, as we need to prevent the case of someone
        # eves-dropping on the password and code and using it again
        # within the 3-minute window. We will leave this to the caller
        # of this function to record!




__all__ = ["QRCodeError", "LoginError", "AccountError",
           "PaymentError", "UserError", "PARError",
           "PARTimeoutError", "PARPermissionsError",
           "PARReadError", "PARWriteError"]


class QRCodeError(Exception):
    pass


class LoginError(Exception):
    pass


class AccountError(Exception):
    pass


class UserError(Exception):
    pass


class PaymentError(Exception):
    pass


class PARError(Exception):
    pass


class PARTimeoutError(PARError):
    pass


class PARPermissionsError(PARError):
    pass


class PARReadError(PARError):
    pass


class PARWriteError(PARError):
    pass




from enum import Enum as _Enum

__all__ = ["TransactionInfo", "TransactionCode"]


class TransactionCode(_Enum):
    CREDIT = "CR"
    DEBIT = "DR"
    CURRENT_LIABILITY = "CL"
    ACCOUNT_RECEIVABLE = "AR"
    RECEIVED_RECEIPT = "RR"
    SENT_RECEIPT = "SR"
    RECEIVED_REFUND = "RF"
    SENT_REFUND = "SF"


class TransactionInfo:
    """This class is used to encode and extract the type of transaction
       and value to/from an object store key
    """
    def __init__(self, key):
        """Extract information from the passed object store key.
           This looks for the string in the key that matches
           '2 letters followed by a number'

           CL000100.005000
           DR000004.234100

           etc.

           For sent and received receipts there are two values;
           the receipted value and the original estimate. These
           have the standard format if the values are the same, e.g.

           RR000100.005000

           however, they have original value T receipted value if they are
           different, e.g.

           RR000100.005000T000090.000000
        """
        from Acquire.Accounting import create_decimal as _create_decimal

        parts = key.split("/")

        # start at the end...
        for i in range(-1, -len(parts), -1):
            part = parts[i]

            try:
                code = TransactionInfo._get_code(part[0:2])

                if code == TransactionCode.SENT_RECEIPT or \
                        code == TransactionCode.RECEIVED_RECEIPT:

                    values = part[2:].split("T")
                    try:
                        value = _create_decimal(values[0])
                        receipted_value = _create_decimal(values[1])
                        self._code = code
                        self._value = value
                        self._receipted_value = receipted_value
                        return
                    except:
                        pass

                value = _create_decimal(part[2:])

                self._code = code
                self._value = value
                self._receipted_value = value
                return
            except:
                pass

        raise ValueError("Cannot extract transaction info from '%s'"
                         % (key))

    def __str__(self):
        return "TransactionInfo(code==%s, value==%s)" % \
                    (self._code.value, self._value)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self._code == other._code and \
                   self._value == other._value
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    @staticmethod
    def _get_code(code):
        """Return the TransactionCode matching 'code'"""
        return TransactionCode(code)

    @staticmethod
    def encode(code, value, receipted_value=None):
        """Encode the passed code and value into a simple string that can
           be used as part of an object store key. If 'receipted_value' is
           passed, then encode the receipted value of the provisional
           transaction too
        """
        if receipted_value is None:
            return "%2s%013.6f" % (code.value, value)
        else:
            return "%2s%013.6fT%013.6f" % (code.value, value, receipted_value)

    def value(self):
        """Return the value of the transaction"""
        return self._value

    def receipted_value(self):
        """Return the receipted value of the transaction. This may be
           different to value() when the transaction was provisional,
           and the receipted value is less than the provisional value
        """
        return self._receipted_value

    def is_credit(self):
        """Return whether or not this is a credit"""
        return self._code == TransactionCode.CREDIT

    def is_debit(self):
        """Return whether or not this is a debit"""
        return self._code == TransactionCode.DEBIT

    def is_liability(self):
        """Return whether or not this is a liability"""
        return self._code == TransactionCode.CURRENT_LIABILITY

    def is_accounts_receivable(self):
        """Return whether or not this is accounts receivable"""
        return self._code == TransactionCode.ACCOUNT_RECEIVABLE

    def is_sent_receipt(self):
        """Return whether or not this is a sent receipt"""
        return self._code == TransactionCode.SENT_RECEIPT

    def is_received_receipt(self):
        """Return whether or not this is a received receipt"""
        return self._code == TransactionCode.RECEIVED_RECEIPT

    def is_sent_refund(self):
        """Return whether or not this is a sent refund"""
        return self._code == TransactionCode.SENT_REFUND

    def is_received_refund(self):
        """Return whether or not this is a received refund"""
        return self._code == TransactionCode.RECEIVED_REFUND




__all__ = ["get_filesize_and_checksum", "get_size_and_checksum"]


def get_size_and_checksum(data):
    """Return a tuple of the size in bytes of the passed data and the
       data's md5 checksum
    """
    from hashlib import md5 as _md5
    md5 = _md5()
    md5.update(data)

    return (len(data), str(md5.hexdigest()))


def get_filesize_and_checksum(filename):
    """Return a tuple of the size in bytes of the passed file and the
       file's md5 checksum
    """
    from hashlib import md5 as _md5
    md5 = _md5()
    size = 0

    with open(filename, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            md5.update(chunk)
            size += len(chunk)

    return (size, str(md5.hexdigest()))




import os as _os
import uuid as _uuid

from Acquire.Access import Request as _Request

__all__ = ["RunRequest"]


def _get_abspath_size_md5(basedir, key, filename, max_size=None):
    """Assert that the specified filename associated with key 'key' exists
       and is readable by this user. Assert also that the filesize if below
       'size' bytes, is 'max_size' has been specified. This returns the
       absolute filename path for the file, the size of the file in bytes
       and the md5 checksum of the file, as a tuple
    """

    if _os.path.isabs(filename):
        filename = _os.path.realpath(filename)
    else:
        filename = _os.path.realpath(_os.path.join(basedir, filename))

    try:
        FILE = open(filename, "r")
        FILE.close()
    except Exception as e:
        from Acquire.Service import exception_to_string
        from Acquire.Access import RunRequestError
        raise RunRequestError(
            "Cannot complete the run request because the file '%s' is not "
            "readable: filename=%s.\n\nCAUSE: %s" %
            (key, filename, exception_to_string(e)))

    from Acquire.Access import get_filesize_and_checksum \
        as _get_filesize_and_checksum

    (filesize, md5) = _get_filesize_and_checksum(filename)

    if filesize > max_size:
        raise RunRequestError(
            "Cannot complete the run request because the file '%s' is "
            "too large: filename=%s, filesize=%f MB, max_size=%f MB" %
            (key, filename, filesize/(1024.0*1024.0),
             max_size/(1024.0*1024.0)))

    return (filename, filesize, md5)


class RunRequest(_Request):
    """This class holds a request to run a particular calculation
       on a RunService. The result of this request will be a
       PAR to which the input should be loaded, and a Bucket
       from which the output can be read. The calculation will
       start once the input has been signalled as loaded.
    """
    def __init__(self, runfile=None):
        """Construct the request
        """
        super().__init__()

        self._uid = None
        self._runinfo = None
        self._tarfile = None
        self._tarfilename = None
        self._tarsize = None
        self._tarmd5 = None

        if runfile is not None:
            # Package up the simulation described in runfile
            self._set_runfile(runfile)

    def is_null(self):
        """Return whether or not this is a null request"""
        return self._uid is None

    def __str__(self):
        if self.is_null():
            return "RunRequest::null"
        else:
            return "RunRequest(uid=%s)" % self._uid

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self._uid == other._uid
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def uid(self):
        """Return the UID of this request"""
        return self._uid

    def fingerprint(self):
        """Return a unique fingerprint for this request that can be
           used for signing and verifying authorisations
        """
        if self.is_null():
            return None

        return "%s%s%s" % (self.uid(), self.tarfile_size(),
                           self.tarfile_md5sum())

    def tarfile(self):
        """Return the name of the tarfile containing all of the
           input files
        """
        return self._tarfilename

    def tarfile_size(self):
        """Return the size of the tarfile in bytes"""
        return self._tarsize

    def tarfile_md5sum(self):
        """Return the MD5 checksum of the tarfile containing
           the input files"""
        return self._tarmd5

    def runinfo(self):
        """Return the processed run information used to describe the
           calculation to be run. This includes information about all
           of the input files, such as their names, filesizes and
           MD5 checksums
        """
        import copy as _copy
        return _copy.deepcopy(self._runinfo)

    def input_files(self):
        """Return a dictionary of the input file information for the
           input files for the calculation. This is a dictionary mapping
           the key for each file to the filename in the tarfile, the
           size of the file in the tarfile and the md5 sum of the file
        """
        if self._runinfo is None:
            return None

        if "input" in self._runinfo:
            return self._runinfo["input"]
        else:
            return None

    def _validate_input(self, basedir, runinfo):
        """Validate that the passed input 'runinfo' is correct, given
           it was loaded from the directory 'basedir'. This
           makes sure that all of the input files exist and are readable
           relative to 'basedir'. These MUST be declared in the 'input'
           section of the dictionary. This returns an updated 'runinfo'
           which has all relative paths converted to absolute file paths
        """
        if "input" not in runinfo:
            return runinfo

        try:
            items = runinfo["input"][0].items()
        except:
            try:
                items = runinfo["input"].items()
            except:
                from Acquire.Access import RunRequestError
                raise RunRequestError(
                    "Cannot execute the request because the input files "
                    "are specified with the wrong format. They should be "
                    "a single dictionary of key-value pairs. "
                    "Instead it is '%s'" % str(runinfo["input"]))

        input = {}

        for (key, filename) in items:
            # check the file exists and is not more than 100 MB is size
            (absfile, filesize, md5) = _get_abspath_size_md5(
                                                basedir, key,
                                                filename,
                                                100*1024*1024)

            input[key] = (absfile, filesize, md5)

        runinfo["input"] = input

        return runinfo

    def _create_tarfile(self):
        """This function creates the new tarfile and updates the
           runinfo with the paths for the input files in the zipfile
        """
        if self._tarfile is not None:
            from Acquire.Access import RunRequestError
            raise RunRequestError("You cannot create the tarfile twice...")

        if "input" not in self._runinfo:
            return

        input = self._runinfo["input"]

        import tarfile as _tarfile
        import tempfile as _tempfile

        # loop through each file - add it to tarbz2. The files are added
        # flat into the tar.bz2, i.e. with no subdirectory. This is to
        # prevent strange complications or clashes with other files that
        # the user may create during output (on the server the files will
        # be unpacked into a uniquely-named directory)
        names = {}

        tempfile = _tempfile.NamedTemporaryFile(suffix="tar.bz2")
        tarfile = _tarfile.TarFile(fileobj=tempfile, mode="w")

        for (key, fileinfo) in input.items():
            (filename, filesize, md5) = fileinfo

            name = _os.path.basename(filename)

            # make sure that there isn't a matching file in the tarfile
            i = 0
            while name in names:
                i += 1
                name = "%d_%s" % (i, name)

            tarfile.add(name=filename, arcname=name, recursive=False)

            input[key] = (name, filesize, md5)

        tarfile.close()

        # close the file so that it is written to the disk - if we close
        # the tempfile then the file is deleted... (which shouldn't happen
        # until the object is deleted)
        tempfile.file.close()

        self._tarfile = tempfile
        self._tarfilename = tempfile.name

        from Acquire.Access import get_filesize_and_checksum \
            as _get_filesize_and_checksum
        (filesize, md5) = _get_filesize_and_checksum(tempfile.name)

        self._tarsize = filesize
        self._tarmd5 = md5

    def _set_runfile(self, runfile):
        """Run the simulation described in the passed runfile (should
           be in yaml or json format). This gives the type of simulation, the
           location of the input files and how the output should be
           named
        """
        if self._runinfo:
            from Acquire.Access import RunRequestError
            raise RunRequestError(
                "You cannot change runfile of this RunRequest")

        if runfile is None:
            return

        runlines = None

        try:
            with open(runfile, "r") as FILE:
                runlines = FILE.read()
        except Exception as e:
            from Acquire.Service import exception_to_string
            from Acquire.Access import RunRequestError
            raise RunRequestError(
                "Cannot read '%s'. You must supply a readable input file "
                "that describes the calculation to be performed and supplies "
                "the names of all of the input files.\n\nCAUSE: %s" %
                (runfile, exception_to_string(e)))

        # get the directory that contains this file
        basedir = _os.path.dirname(_os.path.abspath(runfile))

        # try to parse this input as yaml
        runinfo = None

        try:
            import yaml as _yaml
            runinfo = _yaml.safe_load(runlines)
        except:
            pass

        if runinfo is None:
            try:
                import json as _json
                runinfo = _json.loads(runlines)
            except:
                pass

        if runinfo is None:
            from Acquire.Access import RunRequestError
            raise RunRequestError(
                "Cannot interpret valid input read from the file '%s'. "
                "This should be in json or yaml format, and this parser "
                "be built with that support." % runfile)

        runinfo = self._validate_input(basedir, runinfo)
        self._runinfo = runinfo

        self._create_tarfile()

        # everything is ok - set the UID of this request
        self._uid = str(_uuid.uuid4())

    def to_data(self):
        """Return this request as a json-serialisable dictionary"""
        if self.is_null():
            return {}

        data = super().to_data()
        data["uid"] = self._uid
        data["runinfo"] = self._runinfo
        data["tarsize"] = self._tarsize
        data["tarmd5"] = self._tarmd5

        return data

    @staticmethod
    def from_data(data):
        if (data and len(data) > 0):
            r = RunRequest()

            r._uid = data["uid"]
            r._runinfo = data["runinfo"]
            r._tarsize = int(data["tarsize"])
            r._tarmd5 = data["tarmd5"]

            r._from_data(data)

            return r

        return None




import os as _os
import json as _json

from cachetools import cached as _cached
from cachetools import LRUCache as _LRUCache

# The cache can hold a maximum of 50 objects, and will remove the
# least recently used items from the cache
_login_cache = _LRUCache(maxsize=50)

__all__ = ["get_service_account_bucket",
           "push_testing_objstore", "pop_testing_objstore",
           "clear_login_cache"]

_current_testing_objstore = None
_testing_objstore_stack = []


def clear_login_cache():
    """Call to clear the login cache"""
    _login_cache.clear()


def push_testing_objstore(testing_dir):
    """Function used in testing to push a new object store onto the stack"""
    from Acquire.Service import clear_service_cache as _clear_service_cache

    global _current_testing_objstore
    global _testing_objstore_stack

    _testing_objstore_stack.append(_current_testing_objstore)
    _current_testing_objstore = testing_dir
    _clear_service_cache()


def pop_testing_objstore():
    """Function used in testing to pop an object store from the stack"""
    from Acquire.Service import clear_service_cache as _clear_service_cache

    global _current_testing_objstore
    global _testing_objstore_stack

    try:
        d = _testing_objstore_stack.pop()
    except:
        d = None

    _current_testing_objstore = d
    _clear_service_cache()


# Cache this function as the result changes very infrequently, as involves
# lots of round trips to the object store, and it will give the same
# result regardless of which Fn function on the service makes the call
@_cached(_login_cache)
def get_service_account_bucket(testing_dir=None):
    """This function logs into the object store account of the service account.
       Accessing the object store means being able to access
       all resources and which can authorise the creation
       of access all resources on the object store. Obviously this is
       a powerful account, so only log into it if you need it!!!

       The login information should not be put into a public
       repository or stored in plain text. In this case,
       the login information is held in an environment variable
       (which should be encrypted or hidden in some way...)
    """
    from Acquire.Service import assert_running_service as \
        _assert_running_service

    _assert_running_service()

    # read the password for the secret key from the filesystem
    try:
        with open("secret_key", "r") as FILE:
            password = FILE.readline()[0:-1]
    except:
        password = None

        # we must be in testing mode...
        from Acquire.ObjectStore import use_testing_object_store_backend as \
            _use_testing_object_store_backend

        # see if this is running in testing mode...
        global _current_testing_objstore
        if testing_dir:
            _current_testing_objstore = testing_dir
            return _use_testing_object_store_backend(testing_dir)
        elif _current_testing_objstore:
            return _use_testing_object_store_backend(_current_testing_objstore)

    if password is None:
        from Acquire.Service import ServiceAccountError
        raise ServiceAccountError(
            "You need to supply login credentials via the 'secret_key' "
            "file, and 'SECRET_KEY' and 'SECRET_CONFIG' environment "
            "variables! %s" % testing_dir)

    secret_key = _os.getenv("SECRET_KEY")

    if secret_key is None:
        from Acquire.Service import ServiceAccountError
        raise ServiceAccountError(
            "You must supply the password used to unlock the configuration "
            "key in the 'SECRET_KEY' environment variable")

    try:
        secret_key = _json.loads(secret_key)
    except Exception as e:
        from Acquire.Service import ServiceAccountError
        raise ServiceAccountError(
            "Unable to decode valid JSON from the secret key: %s" % str(e))

    # use the password to decrypt the SECRET_KEY in the config
    try:
        from Acquire.Crypto import PrivateKey as _PrivateKey
        secret_key = _PrivateKey.from_data(secret_key, password)
    except Exception as e:
        from Acquire.Service import ServiceAccountError
        raise ServiceAccountError(
            "Unable to open the private SECRET_KEY using the password "
            "supplied in the 'secret_key' file: %s" % str(e))

    config = _os.getenv("SECRET_CONFIG")

    if config is None:
        from Acquire.Service import ServiceAccountError
        raise ServiceAccountError(
            "You must supply the encrypted config in teh 'SECRET_CONFIG' "
            "environment variable!")

    try:
        from Acquire.ObjectStore import string_to_bytes as _string_to_bytes
        config = secret_key.decrypt(_string_to_bytes(config))
    except Exception as e:
        from Acquire.Service import ServiceAccountError
        raise ServiceAccountError(
            "Cannot decrypt the 'SECRET_CONFIG' with the 'SECRET_KEY'. Are "
            "you sure that the configuration has been set up correctly? %s "
            % str(e))

    # use the secret_key to decrypt the config in SECRET_CONFIG
    try:
        config = _json.loads(config)
    except Exception as e:
        from Acquire.Service import ServiceAccountError
        raise ServiceAccountError(
            "Unable to decode valid JSON from the config: %s" % str(e))

    # get info from this config
    access_data = config["LOGIN"]
    bucket_data = config["BUCKET"]

    # save the service password to the environment
    _os.environ["SERVICE_PASSWORD"] = config["PASSWORD"]

    # save any other decrypted config data to environment variables
    for key in config.keys():
        if key not in ["LOGIN", "BUCKET", "PASSWORD"]:
            _os.environ[key] = config[key]

    # we have OCI login details, so make sure that we are using
    # the OCI object store backend
    from Acquire.ObjectStore import use_oci_object_store_backend as \
        _use_oci_object_store_backend

    _use_oci_object_store_backend()

    # now login and create/load the bucket for this account
    try:
        from ._oci_account import OCIAccount as _OCIAccount

        account_bucket = _OCIAccount.create_and_connect_to_bucket(
                                    access_data,
                                    bucket_data["compartment"],
                                    bucket_data["bucket"])
    except Exception as e:
        from Acquire.Service import ServiceAccountError
        raise ServiceAccountError(
             "Error connecting to the service account: %s" % str(e))

    return account_bucket




import io as _io
import datetime as _datetime
import uuid as _uuid
import json as _json
import os as _os
import copy as _copy
import uuid as _uuid

__all__ = ["OCI_ObjectStore"]


def _sanitise_bucket_name(bucket_name):
    """This function sanitises the passed bucket name. It will always
        return a valid bucket name. If "None" is passed, then a new,
        unique bucket name will be generated"""

    if bucket_name is None:
        return str(_uuid.uuid4())

    return "_".join(bucket_name.split())


def _clean_key(key):
    """This function cleans and returns a key so that it is suitable
       for use both as a key and a directory/file path
       e.g. it removes double-slashes
    """
    return _os.path.normpath(key)


def _get_object_url_for_region(region, uri):
    """Internal function used to get the full URL to the passed PAR URI
       for the specified region. This has the format;

       https://objectstorage.{region}.oraclecloud.com/{uri}
    """
    server = "https://objectstorage.%s.oraclecloud.com" % region

    while uri.startswith("/"):
        uri = uri[1:]

    return "%s/%s" % (server, uri)


class OCI_ObjectStore:
    """This is the backend that abstracts using the Oracle Cloud
       Infrastructure object store
    """

    @staticmethod
    def create_bucket(bucket, bucket_name, compartment=None):
        """Create and return a new bucket in the object store called
           'bucket_name', optionally placing it into the compartment
           identified by 'compartment'. This will raise an
           ObjectStoreError if this bucket already exists
        """
        new_bucket = _copy.copy(bucket)

        new_bucket["bucket_name"] = str(bucket_name)

        if compartment is not None:
            new_bucket["compartment_id"] = str(compartment)

        try:
            from oci.object_storage.models import CreateBucketDetails as \
                _CreateBucketDetails
        except:
            raise ImportError(
                "Cannot import OCI. Please install OCI, e.g. via "
                "'pip install oci' so that you can connect to the "
                "Oracle Cloud Infrastructure")

        try:
            request = _CreateBucketDetails()
            request.compartment_id = new_bucket["compartment_id"]
            client = new_bucket["client"]
            request.name = _sanitise_bucket_name(bucket_name)

            new_bucket["bucket"] = client.create_bucket(
                                        client.get_namespace().data,
                                        request).data
        except Exception as e:
            # couldn't create the bucket - likely because it already
            # exists - try to connect to the existing bucket
            from Acquire.ObjectStore import ObjectStoreError
            raise ObjectStoreError(
                "Unable to create the bucket '%s', likely because it "
                "already exists: %s" % (bucket_name, str(e)))

        return new_bucket

    @staticmethod
    def get_bucket(bucket, bucket_name, compartment=None,
                   create_if_needed=True):
        """Find and return a new bucket in the object store called
           'bucket_name', optionally placing it into the compartment
           identified by 'compartment'. If 'create_if_needed' is True
           then the bucket will be created if it doesn't exist. Otherwise,
           if the bucket does not exist then an exception will be raised.
        """
        new_bucket = _copy.copy(bucket)

        new_bucket["bucket_name"] = _sanitise_bucket_name(bucket_name)

        if compartment is not None:
            new_bucket["compartment_id"] = str(compartment)

        try:
            from oci.object_storage.models import CreateBucketDetails as \
                _CreateBucketDetails
        except:
            raise ImportError(
                "Cannot import OCI. Please install OCI, e.g. via "
                "'pip install oci' so that you can connect to the "
                "Oracle Cloud Infrastructure")

        # try to get the existing bucket
        client = new_bucket["client"]
        namespace = client.get_namespace().data
        sanitised_name = _sanitise_bucket_name(bucket_name)

        try:
            existing_bucket = client.get_bucket(
                                namespace, sanitised_name).data
        except:
            existing_bucket = None

        if existing_bucket:
            new_bucket["bucket"] = existing_bucket
            return new_bucket

        if create_if_needed:
            try:
                request = _CreateBucketDetails()
                request.compartment_id = new_bucket["compartment_id"]
                request.name = sanitised_name

                client.create_bucket(namespace, request)
            except:
                pass

            try:
                existing_bucket = client.get_bucket(
                                    namespace, sanitised_name).data
            except:
                existing_bucket = None

        if existing_bucket is None:
            from Acquire.ObjectStore import ObjectStoreError
            raise ObjectStoreError(
                "There is not bucket called '%s'. Please check the "
                "compartment and access permissions." % bucket_name)

        new_bucket["bucket"] = existing_bucket

        return new_bucket

    @staticmethod
    def create_par(bucket, encrypt_key, key=None, readable=True,
                   writeable=False, duration=3600):
        """Create a pre-authenticated request for the passed bucket and
           key (if key is None then the request is for the entire bucket).
           This will return a PAR object that will contain a URL that can
           be used to access the object/bucket. If writeable is true, then
           the URL will also allow the object/bucket to be written to.
           PARs are time-limited. Set the lifetime in seconds by passing
           in 'duration' (by default this is one hour)
        """
        from Acquire.Crypto import PublicKey as _PublicKey

        if not isinstance(encrypt_key, _PublicKey):
            from Acqure.Client import PARError
            raise PARError(
                "You must supply a valid PublicKey to encrypt the "
                "returned PAR")

        # get the UTC datetime when this PAR should expire
        from Acquire.ObjectStore import get_datetime_now as _get_datetime_now
        expires_datetime = _get_datetime_now() + \
            _datetime.timedelta(seconds=duration)

        is_bucket = (key is None)

        # Limitation of OCI - cannot have a bucket PAR with
        # read permissions!
        if is_bucket and readable:
            from Acquire.Client import PARError
            raise PARError(
                "You cannot create a Bucket PAR that has read permissions "
                "due to a limitation in the underlying platform")

        try:
            from oci.object_storage.models import \
                CreatePreauthenticatedRequestDetails as \
                _CreatePreauthenticatedRequestDetails
        except:
            raise ImportError(
                "Cannot import OCI. Please install OCI, e.g. via "
                "'pip install oci' so that you can connect to the "
                "Oracle Cloud Infrastructure")

        oci_par = None

        try:
            request = _CreatePreauthenticatedRequestDetails()

            if is_bucket:
                request.access_type = "AnyObjectWrite"
            elif readable and writeable:
                request.access_type = "ObjectReadWrite"
            elif readable:
                request.access_type = "ObjectRead"
            elif writeable:
                request.access_type = "ObjectWrite"
            else:
                from Acquire.ObjectStore import ObjectStoreError
                raise ObjectStoreError(
                    "Unsupported permissions model for PAR!")

            request.name = str(_uuid.uuid4())

            if not is_bucket:
                request.object_name = _clean_key(key)

            request.time_expires = expires_datetime

            client = bucket["client"]

            oci_par = client.create_preauthenticated_request(
                                        client.get_namespace().data,
                                        bucket["bucket_name"],
                                        request).data
        except Exception as e:
            # couldn't create the preauthenticated request
            from Acquire.ObjectStore import ObjectStoreError
            raise ObjectStoreError(
                "Unable to create the preauthenticated request '%s': %s" %
                (str(request), str(e)))

        if oci_par is None:
            from Acquire.ObjectStore import ObjectStoreError
            raise ObjectStoreError(
                "Unable to create the preauthenticated request!")

        created_datetime = oci_par.time_created.replace(
                                tzinfo=_datetime.timezone.utc)

        expires_datetime = oci_par.time_expires.replace(
                                tzinfo=_datetime.timezone.utc)

        # the URI returned by OCI does not include the server. We need
        # to get the server based on the region of this bucket
        url = _get_object_url_for_region(bucket["region"],
                                         oci_par.access_uri)

        from Acquire.Client import PAR as _PAR
        par = _PAR(url=url, encrypt_key=encrypt_key,
                   key=oci_par.object_name,
                   created_datetime=created_datetime,
                   expires_datetime=expires_datetime,
                   is_readable=readable,
                   is_writeable=writeable,
                   par_id=str(oci_par.id),
                   par_name=str(oci_par.name),
                   driver="oci")

        return par

    @staticmethod
    def get_object_as_file(bucket, key, filename):
        """Get the object contained in the key 'key' in the passed 'bucket'
           and writing this to the file called 'filename'"""

        key = _clean_key(key)

        try:
            response = bucket["client"].get_object(bucket["namespace"],
                                                   bucket["bucket_name"],
                                                   key)
            is_chunked = False
        except:
            try:
                response = bucket["client"].get_object(bucket["namespace"],
                                                       bucket["bucket_name"],
                                                       "%s/1" % key)
                is_chunked = True
            except:
                is_chunked = False
                pass

            if not is_chunked:
                from Acquire.ObjectStore import ObjectStoreError
                raise ObjectStoreError("No object at key '%s'" % key)

        if not is_chunked:
            with open(filename, 'wb') as f:
                for chunk in response.data.raw.stream(1024 * 1024,
                                                      decode_content=False):
                    f.write(chunk)

            return filename

        # the data is chunked - get this out chunk by chunk
        with open(filename, 'wb') as f:
            next_chunk = 1
            while True:
                for chunk in response.data.raw.stream(1024 * 1024,
                                                      decode_content=False):
                    f.write(chunk)

                # now get and write the rest of the chunks
                next_chunk += 1

                try:
                    response = bucket["client"].get_object(
                        bucket["namespace"], bucket["bucket_name"],
                        "%s/%d" % (key, next_chunk))
                except:
                    break

    @staticmethod
    def get_object(bucket, key):
        """Return the binary data contained in the key 'key' in the
           passed bucket"""

        key = _clean_key(key)

        try:
            response = bucket["client"].get_object(bucket["namespace"],
                                                   bucket["bucket_name"],
                                                   key)
            is_chunked = False
        except:
            try:
                response = bucket["client"].get_object(bucket["namespace"],
                                                       bucket["bucket_name"],
                                                       "%s/1" % key)
                is_chunked = True
            except:
                is_chunked = False
                pass

            # Raise the original error
            if not is_chunked:
                from Acquire.ObjectStore import ObjectStoreError
                raise ObjectStoreError("No data at key '%s'" % key)

        data = None

        for chunk in response.data.raw.stream(1024 * 1024,
                                              decode_content=False):
            if not data:
                data = chunk
            else:
                data += chunk

        if is_chunked:
            # keep going through to find more chunks
            next_chunk = 1

            while True:
                next_chunk += 1

                try:
                    response = bucket["client"].get_object(
                                        bucket["namespace"],
                                        bucket["bucket_name"],
                                        "%s/%d" % (key, next_chunk))
                except:
                    response = None
                    break

                for chunk in response.data.raw.stream(1024 * 1024,
                                                      decode_content=False):
                    if not data:
                        data = chunk
                    else:
                        data += chunk

        return data

    @staticmethod
    def get_string_object(bucket, key):
        """Return the string in 'bucket' associated with 'key'"""
        key = _clean_key(key)
        return OCI_ObjectStore.get_object(bucket, key).decode("utf-8")

    @staticmethod
    def get_object_from_json(bucket, key):
        """Return an object constructed from json stored at 'key' in
           the passed bucket. This returns None if there is no data
           at this key
        """
        data = None

        key = _clean_key(key)

        try:
            data = OCI_ObjectStore.get_string_object(bucket, key)
        except:
            return None

        return _json.loads(data)

    @staticmethod
    def get_all_object_names(bucket, prefix=None):
        """Returns the names of all objects in the passed bucket"""
        if prefix is not None:
            prefix = _clean_key(prefix)

        objects = bucket["client"].list_objects(bucket["namespace"],
                                                bucket["bucket_name"],
                                                prefix=prefix).data

        names = []

        for obj in objects.objects:
            if prefix:
                if obj.name.startswith(prefix):
                    name = obj.name
            else:
                name = obj.name

            while name.endswith("/"):
                name = name[0:-1]

            while name.startswith("/"):
                name = name[1:]

            if len(name) > 0:
                names.append(name)

        return names

    @staticmethod
    def get_all_objects(bucket, prefix=None):
        """Return all of the objects in the passed bucket"""
        if prefix is not None:
            prefix = _clean_key(prefix)

        objects = {}
        names = OCI_ObjectStore.get_all_object_names(bucket, prefix)

        for name in names:
            objects[name] = OCI_ObjectStore.get_object(bucket, name)

        return objects

    @staticmethod
    def get_all_objects_from_json(bucket, prefix=None):
        """Return all of the json objects in the passed bucket as
           json-deserialised objects
        """
        objects = OCI_ObjectStore.get_all_objects(bucket, prefix)

        names = list(objects.keys())

        for name in names:
            try:
                s = objects[name].decode("utf-8")
                objects[name] = _json.loads(s)
            except:
                del objects[name]

        return objects

    @staticmethod
    def get_all_strings(bucket, prefix=None):
        """Return all of the strings in the passed bucket"""
        objects = OCI_ObjectStore.get_all_objects(bucket, prefix)

        names = list(objects.keys())

        for name in names:
            try:
                s = objects[name].decode("utf-8")
                objects[name] = s
            except:
                del objects[name]

        return objects

    @staticmethod
    def set_object(bucket, key, data):
        """Set the value of 'key' in 'bucket' to binary 'data'"""
        f = _io.BytesIO(data)

        key = _clean_key(key)
        bucket["client"].put_object(bucket["namespace"],
                                    bucket["bucket_name"],
                                    key, f)

    @staticmethod
    def set_object_from_file(bucket, key, filename):
        """Set the value of 'key' in 'bucket' to equal the contents
           of the file located by 'filename'"""
        key = _clean_key(key)
        with open(filename, 'rb') as f:
            bucket["client"].put_object(bucket["namespace"],
                                        bucket["bucket_name"],
                                        key, f)

    @staticmethod
    def set_string_object(bucket, key, string_data):
        """Set the value of 'key' in 'bucket' to the string 'string_data'"""
        key = _clean_key(key)
        OCI_ObjectStore.set_object(bucket, key, string_data.encode("utf-8"))

    @staticmethod
    def set_object_from_json(bucket, key, data):
        """Set the value of 'key' in 'bucket' to equal to contents
           of 'data', which has been encoded to json"""
        key = _clean_key(key)
        OCI_ObjectStore.set_string_object(bucket, key, _json.dumps(data))

    @staticmethod
    def delete_all_objects(bucket, prefix=None):
        """Deletes all objects..."""

        for obj in OCI_ObjectStore.get_all_object_names(bucket):
            bucket["client"].delete_object(bucket["namespace"],
                                           bucket["bucket_name"],
                                           obj)

    @staticmethod
    def delete_object(bucket, key):
        """Removes the object at 'key'"""
        try:
            key = _clean_key(key)
            bucket["client"].delete_object(bucket["namespace"],
                                           bucket["bucket_name"],
                                           key)
        except:
            pass

    @staticmethod
    def clear_all_except(bucket, keys):
        """Removes all objects from the passed 'bucket' except those
           whose keys are or start with any key in 'keys'"""
        names = OCI_ObjectStore.get_all_object_names(bucket)

        for name in names:
            remove = True

            for key in keys:
                key = _clean_key(key)
                if name.startswith(key):
                    remove = False
                    break

            if remove:
                bucket["client"].delete_object(bucket["namespace"],
                                               bucket["bucket_name"],
                                               name)




from uuid import uuid4 as _uuid4
import copy as _copy
import os as _os

from Acquire.Access import Request as _Request

__all__ = ["FileWriteRequest"]


def _get_key(root_key, filename):
    """Return the object store key for 'filename', using 'root_key' as
       the root
    """
    basename = _os.path.basename(filename)

    if root_key:
        return "%s/%s" % (str(root_key), basename)
    else:
        return basename


def _clean_key(root_key, filekey):
    """Return the cleaned key 'filekey', using 'root_key' as the root"""
    if root_key:
        return "%s/%s" % (str(root_key), str(filekey))
    else:
        return str(filekey)


def _get_filesize_and_checksum(filename):
    """Return a tuple of the size in bytes of the passed file and the
       file's md5 checksum
    """
    from hashlib import md5 as _md5
    md5 = _md5()
    size = 0

    with open(filename, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            md5.update(chunk)
            size += len(chunk)

    return (size, str(md5.hexdigest()))


def _list_all_files(directory, ignore_hidden=True):
    """Return a list of the path relative to 'directory' of
       all files contained in 'directory'. If is_hidden is True, then include
       all hidden files - otherwise these are ignored
    """
    if not _os.path.isdir(directory):
        return []

    all_files = []

    for (root, _, filenames) in _os.walk(directory):
        root = root[len(directory)+1:]

        if ignore_hidden and root.startswith("."):
            continue

        for filename in filenames:
            if not (filename.startswith(".") and ignore_hidden):
                all_files.append(_os.path.join(root, filename))

    return all_files


def _clean_path(path):
    """This function cleans the passed path so that doesn't contain
       redundant slashes or '..' etc., so that all backslashes are forwards
       slashes, and that the trailing slash is removed
    """
    if path is None:
        return ""

    path = _os.path.normpath(path)

    # remove all ".." and "." from this path
    if path.find(".") != -1:
        parts = path.split("/")
        for i, part in enumerate(parts):
            if part == "." or part == "..":
                parts[i] = ""

        path = _os.path.normpath("/".join(parts))

    return path


def _expand_source_destination(source, destination=None,
                               root=None, ignore_hidden=True):
    """This function expands the 'source' and 'destination' into a pair
       of lists - the source files and the destination keys in the
       object store.
    """
    if source is None:
        return ([], [])

    if not isinstance(source, list):
        source = [str(source)]

    is_destination_dir = False
    try:
        is_destination_dir = destination.endswith("/")
    except:
        pass

    destination = _os.path.normpath(_os.path.join(_clean_path(root),
                                                  _clean_path(destination)))

    # expand all of the sources into the full paths of files (which must exist)
    source_filenames = []
    rel_filenames = []

    from glob import glob as _glob

    for s in source:
        for f in _glob(str(s)):
            abspath = _os.path.abspath(f)
            if not _os.path.exists(abspath):
                raise FileExistsError("The file '%s' does not exist!"
                                      % abspath)

            if _os.path.isdir(abspath):
                dirfiles = _list_all_files(abspath, ignore_hidden)

                for dirfile in dirfiles:
                    source_filenames.append(_os.path.join(abspath, dirfile))
                    rel_filenames.append(dirfile)
            else:
                source_filenames.append(abspath)
                rel_filenames.append(_os.path.basename(abspath))

    destination_keys = []

    if len(rel_filenames) == 1:
        if is_destination_dir:
            destination_keys.append(_os.path.join(destination,
                                                  rel_filenames[0]))
        else:
            destination_keys.append(destination)
    else:
        for rel_filename in rel_filenames:
            destination_keys.append(_os.path.join(destination, rel_filename))

    if len(source_filenames) != len(destination_keys):
        raise ValueError("Something went wrong as the number of source "
                         "filenames should be equal to the number of keys")

    return (source_filenames, destination_keys)


class FileWriteRequest(_Request):
    """This class holds a request to write a file (or files) to a remote
       object store
    """
    def __init__(self, source=None, destination=None, root=None,
                 ignore_hidden=True, account=None, testing_key=None):
        """Construct the request to write the files specified in 'source'
           to 'destination' in the cloud object store. You can optionally
           pass a 'root' for all of the keys in the object store, and,
           if the source is a directory, you can ignore hidden files using
           'ignore_hidden=True'.

           You must pass the 'account' from which payment will be taken to
           write files to the object store.
        """
        super().__init__()

        (filenames, keys) = FileWriteRequest.expand_source_and_destination(
                                                source, destination,
                                                root, ignore_hidden)

        self._destination_keys = keys
        self._source_filenames = filenames
        self._file_sizes = []
        self._checksums = []

        if len(self._source_filenames) == 0:
            self._uid = None
            return

        for filename in self._source_filenames:
            (size, checksum) = _get_filesize_and_checksum(filename)
            self._file_sizes.append(size)
            self._checksums.append(checksum)

        self._is_testing = False

        if testing_key:
            self._is_testing = True
            self._testing_key = testing_key
        elif account is not None:
            from Acquire.Client import Account as _Account
            if not isinstance(account, _Account):
                raise TypeError("The passed account must be type Account")

            if not account.is_logged_in():
                raise PermissionError(
                    "You can only authorise payment from the account if you "
                    "have logged in")

        else:
            raise ValueError("You must pass a valid account to write files!")

        self._uid = str(_uuid4())

        from Acquire.Identity import Authorisation as _Authorisation

        if self._is_testing:
            self._account_uid = "something"
            self._accounting_service_url = "somewhere"
            self._authorisation = _Authorisation(
                                    resource=self.resource_key(),
                                    testing_key=self._testing_key)

        else:
            self._account_uid = account.uid()
            self._accounting_service_url = account.accounting_service()\
                                                  .canonical_url()

            self._authorisation = _Authorisation(
                                    resource=self.resource_key(),
                                    user=account.owner())

    def is_null(self):
        """Return whether or not this is a null request"""
        return self._uid is None

    def __str__(self):
        if self.is_null():
            return "FileWriteRequest::null"
        else:
            return "FileWriteRequest(uid=%s)" % self._uid

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self._uid == other._uid
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def resource_key(self):
        """Function to return a string that can be used as a
           summary key for this resource request
        """
        if self.is_null():
            return None
        else:
            return "%s %s" % (self._uid, ":".join(self._checksums))

    def uid(self):
        """Return the UID of this request"""
        return self._uid

    def authorisation(self):
        """Return the authorisation behind this request"""
        return self._authorisation

    def source_filenames(self):
        """Return the filenames of the files to be copied. Note that this
           information is only available in the copy of the object that
           created the request - it is not saved when this object is
           serialised to json as we don't want to leak potentially
           sensitive data to the object store
        """
        return self._source_filenames

    def destination_keys(self):
        """Return the object store keys to which the files will be
           written
        """
        return _copy.copy(self._destination_keys)

    def filesizes(self):
        """Return the sizes of the files that are requested to be written"""
        return _copy.copy(self._file_sizes)

    def checksums(self):
        """Return the checksums of the files that are requested
           to be written"""
        return _copy.copy(self._checksums)

    def account_uid(self):
        """Return the UID of the account from which payment should be
           taken for the file storage
        """
        try:
            return self._account_uid
        except:
            return None

    @staticmethod
    def expand_source_and_destination(source, destination, root=None,
                                      ignore_hidden=True):
        """This function expands the passed source and destination objects
           into the list of absolute paths of local source files, and the
           full keys of those files once they are uploaded to the object
           store. This returns a pair of lists - the lists match the
           absolute path of the local file to the desired full key
           of the file in the object store
        """
        return _expand_source_destination(source, destination, root,
                                          ignore_hidden)

    def accounting_service_url(self):
        """Return the canonical URL of the service holding the account"""
        try:
            return self._accounting_service_url
        except:
            return None

    def to_data(self):
        """Return this request as a json-serialisable dictionary"""
        if self.is_null():
            return {}

        data = super().to_data()
        data["uid"] = self._uid
        data["destination_keys"] = self._destination_keys
        data["file_sizes"] = self._file_sizes
        data["checksums"] = self._checksums
        data["authorisation"] = self._authorisation.to_data()
        data["account_uid"] = self._account_uid
        data["accounting_service_url"] = self._accounting_service_url

        # don't write self._source_filenames as this contains
        # potentially sensitive local data that is not needed
        # by the cloud object store

        return data

    @staticmethod
    def from_data(data):
        if (data and len(data) > 0):
            from Acquire.Identity import Authorisation as _Authorisation
            f = FileWriteRequest()

            f._uid = data["uid"]
            f._destination_keys = data["destination_keys"]
            f._file_sizes = data["file_sizes"]
            f._checksums = data["checksums"]
            f._authorisation = _Authorisation.from_data(data["authorisation"])
            f._account_uid = data["account_uid"]
            f._accounting_service_url = data["accounting_service_url"]

            f._from_data(data)

            return f
        else:
            return None




import uuid as _uuid

__all__ = ["BucketHandle"]


class BucketHandle:
    """This class represents a handle to a bucket in the object
       store. This handle can be used to generate pre-authenticated
       request URLs (PARs - also called pre-signed requests).

       Every bucket has a unique UID. It is up to you to map a
       UID to one or more meaningful names
    """
    def __init__(self, uid=None, compartment=None):
        """Construct a handle to a bucket. If the UID is None then
           this creates a new bucket with a new UID. Otherwise this
           will attempt to connect to the bucket with passed UID.

           By default this will use the same compartment as the
           service bucket. If you want to specify a different
           compartment then pass this in as 'compartment'
        """
        from Acquire.Service import get_service_account_bucket \
            as _get_service_account_bucket
        from Acquire.ObjectStore import ObjectStore as _ObjectStore

        bucket = _get_service_account_bucket()

        if uid is None:
            # try to create a new bucket - attempt to create it three times.
            # If it fails after 3 times then raise the error why
            errors = []

            for _attempt in range(0, 3):
                uid = _uuid.uuid4()
                try:
                    # this call fails if the bucket already exists
                    self._bucket = _ObjectStore.create_bucket(bucket, uid,
                                                              compartment)
                    self._uid = uid
                    return
                except Exception as e:
                    errors.append(e)

            raise errors[-1]
        else:
            self._bucket = _ObjectStore.get_bucket(bucket, uid, compartment,
                                                   create_if_needed=False)
            self._uid = None

    def uid(self):
        """Return the UID of this bucket"""
        return self._uid

    def get_par(self, writeable=False, valid_duration=3600):
        """Return a pre-authenticated request result (PAR) to allow access
           to this bucket. The returned PAR will be valid for 'valid_duration'
           seconds from granting the request. By default PARs are valid for
           one hour
        """
        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        return _ObjectStore.create_par(self._bucket, writeable, valid_duration)




__all__ = ["Accounts"]


class Accounts:
    """This class provides the interface to grouping and ungrouping
       accounts, and associating them with users and services. An account
       can belong to multiple groups. Authorisation of access to accounts
       is based on which group they are in. Typically the groups will
       refer to users, e.g. all of the accounts for a user will be in the
       user's group, and can then be authorised by authorising the user
    """
    def __init__(self, group=None):
        """Construct an interface to the group of accounts called 'group'.
           If the group is not specified, then it will default to 'default'
        """
        if group is None:
            group = "default"

        self._group = str(group)

    def __str__(self):
        return "Accounts(group=%s)" % self.group()

    def _root(self):
        """Return the root key for this group in the object store"""
        from Acquire.ObjectStore import string_to_encoded \
            as _string_to_encoded
        return "accounting/account_groups/%s" % \
            _string_to_encoded(self._group)

    def _account_key(self, name):
        """Return the key for the account called 'name' in this group"""
        from Acquire.ObjectStore import string_to_encoded \
            as _string_to_encoded
        return "%s/%s" % (self._root(),
                          _string_to_encoded(str(name)))

    def group(self):
        """Return the name of the group that this set of accounts refers to"""
        return self._group

    def list_accounts(self, bucket=None):
        """Return the names of all of the accounts in this group"""
        if bucket is None:
            from Acquire.Service import get_service_account_bucket \
                as _get_service_account_bucket
            bucket = _get_service_account_bucket()

        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.ObjectStore import encoded_to_string \
            as _encoded_to_string

        keys = _ObjectStore.get_all_object_names(bucket, self._root())
        root_len = len(self._root())

        accounts = []

        for key in keys:
            try:
                account_key = key[root_len:]
                while account_key.startswith("/"):
                    account_key = account_key[1:]

                accounts.append(_encoded_to_string(account_key))
            except Exception as e:
                from Acquire.Accounting import AccountError
                raise AccountError(
                    "Unable to identify the account associated with key "
                    "'%s', equals '%s': %s" %
                    (key, account_key, str(e)))

        return accounts

    def get_account(self, name, bucket=None):
        """Return the account called 'name' from this group"""
        if bucket is None:
            from Acquire.Service import get_service_account_bucket \
                as _get_service_account_bucket
            bucket = _get_service_account_bucket()

        try:
            from Acquire.ObjectStore import ObjectStore as _ObjectStore
            account_uid = _ObjectStore.get_string_object(
                            bucket, self._account_key(name))
        except:
            account_uid = None

        if account_uid is None:
            # ensure that the user always has a "main" account
            if name == "main":
                return self.create_account("main", "primary user account",
                                           overdraft_limit=0, bucket=bucket)

            from Acquire.Accounting import AccountError
            raise AccountError("There is no account called '%s' in the "
                               "group '%s'" % (name, self.group()))

        from Acquire.Accounting import Account as _Account
        return _Account(uid=account_uid, bucket=bucket)

    def contains(self, account, bucket=None):
        """Return whether or not this group contains the passed account"""
        from Acquire.Accounting import Account as _Account
        if not isinstance(account, _Account):
            raise TypeError("The passed account must be of type Account")

        if bucket is None:
            from Acquire.Service import get_service_account_bucket \
                as _get_service_account_bucket
            bucket = _get_service_account_bucket()

        # read the UID of the account in this group that matches the
        # passed account's name
        try:
            from Acquire.ObjectStore import ObjectStore as _ObjectStore
            account_uid = _ObjectStore.get_string_object(
                            bucket, self._account_key(account.name()))
        except:
            account_uid = None

        return account.uid() == account_uid

    def create_account(self, name, description=None,
                       overdraft_limit=None, bucket=None):
        """Create a new account called 'name' in this group. This will
           return the existing account if it already exists
        """
        if name is None:
            raise ValueError("You must pass a name of the new account")

        account_key = self._account_key(name)

        if bucket is None:
            from Acquire.Service import get_service_account_bucket \
                as _get_service_account_bucket
            bucket = _get_service_account_bucket()

        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.Accounting import Account as _Account

        try:
            account_uid = _ObjectStore.get_string_object(bucket, account_key)
        except:
            account_uid = None

        if account_uid is not None:
            # this account already exists - just return it
            account = _Account(uid=account_uid, bucket=bucket)

            if overdraft_limit is not None:
                account.set_overdraft_limit(overdraft_limit, bucket=bucket)

            return account

        # make sure that no-one has created this account before
        from Acquire.ObjectStore import Mutex as _Mutex
        m = _Mutex(account_key, timeout=600, lease_time=600, bucket=bucket)

        try:
            account_uid = _ObjectStore.get_string_object(bucket, account_key)
        except:
            account_uid = None

        if account_uid is not None:
            m.unlock()
            # this account already exists - just return it
            account = _Account(uid=account_uid, bucket=bucket)

            if overdraft_limit is not None:
                account.set_overdraft_limit(overdraft_limit, bucket=bucket)

            return account

        # write a temporary UID to the object store so that we
        # can ensure we are the only function to create it
        try:
            _ObjectStore.set_string_object(bucket, account_key,
                                           "under_construction")
        except:
            m.unlock()
            raise

        m.unlock()

        # ok - we are the only function creating this account. Let's try
        # to create it properly
        try:
            account = _Account(name=name, description=description,
                               bucket=bucket)
        except:
            try:
                _ObjectStore.delete_object(bucket, account_key)
            except:
                pass

            raise

        if overdraft_limit is not None:
            account.set_overdraft_limit(overdraft_limit, bucket=bucket)

        _ObjectStore.set_string_object(bucket, account_key, account.uid())

        return account



"""
Acquire : (C) Christopher Woods 2018

This module provides the base classes and functions used to define
the services used in the system. It is not likely to be user-facing
"""

from ._function import *
from ._get_users_public_certs import *
from ._get_services import *
from ._get_service_account_bucket import *
from ._service_account import *
from ._service import *
from ._profile import *
from ._errors import *
from ._cache_management import *
from ._trust_service import *

try:
    if __IPYTHON__:
        def _set_printer(C):
            """Function to tell ipython to use __str__ if available"""
            get_ipython().display_formatter.formatters['text/plain'].for_type(
                C,
                lambda obj, p, cycle: p.text(str(obj) if not cycle else '...')
                )

        import sys as _sys
        import inspect as _inspect

        _clsmembers = _inspect.getmembers(_sys.modules[__name__],
                                          _inspect.isclass)

        for _clsmember in _clsmembers:
            _set_printer(_clsmember[1])
except:
    pass




__all__ = ["Transaction"]


def _getcontext():
    """Return the context used for all decimals in transactions. This
       context rounds to 6 decimal places and provides sufficient precision
       to support any value between 0 and 999,999,999,999,999.999,999,999
       (i.e. everything up to just under one quadrillion - I doubt we will
        ever have an account that has more than a trillion units in it!)
    """
    from decimal import Context as _Context
    return _Context(prec=24)


def _create_decimal(value):
    """Create a decimal from the passed value. This is a decimal that
       has 6 decimal places and is clamped between 0 <= value < 1 quadrillion
    """
    from decimal import Decimal as _Decimal

    try:
        d = _Decimal("%.6f" % value, _getcontext())
    except:
        value = _Decimal(value, _getcontext())
        d = _Decimal("%.6f" % value, _getcontext())

    if d < 0:
        from Acquire.Accounting import TransactionError
        raise TransactionError(
                "You cannot create a transaction with a negative value (%s)"
                % (value))

    elif d >= 1000000000000000:
        from Acquire.Accounting import TransactionError
        raise TransactionError(
                "You cannot create a transaction with a value greater than "
                "1 quadrillion! (%s)" % (value))

    return d


class Transaction:
    """This class provides basic information about a transaction - namely
       just the value (always positive) and what the transaction is for
    """
    def __init__(self, value=0, description=None):
        """Create a transaction with the passed value and description. Values
           are positive values with a minimum resolution of 1e-6 (6 decimal
           places) and cannot exceed 999999.999999 (i.e. cannot be 1 million
           or greater). This is to ensure that a value will always fit into a
           f13.6 string (which is used for keys). If you need larger
           transactions, then use the static 'split' function to break
           a single too-large transaction into a list of smaller
           transactions
        """
        value = _create_decimal(value)

        if value > Transaction.maximum_transaction_value():
            from Acquire.Accounting import TransactionError
            raise TransactionError(
                "You cannot create a transaction (%s) with a "
                "value greater than %s. Please "
                "use 'split' to break this transaction (%s) into "
                "several separate transactions" %
                (description, Transaction.maximum_transaction_value(), value))

        # ensure that the value is limited in resolution to 6 decimal places
        self._value = value

        if self._value < 0:
            from Acquire.Accounting import TransactionError
            raise TransactionError(
                "You cannot create a transaction (%s) with a "
                "negative value! %s" % (description, value))

        if description is None:
            if self._value > 0:
                from Acquire.Accounting import TransactionError
                raise TransactionError(
                    "You must give a description to all non-zero "
                    "transactions! %s" % self.value())
            else:
                self._description = None
        else:
            # ensure we are using utf-8 encoded strings
            self._description = str(description).encode("utf-8") \
                                                .decode("utf-8")

    def __str__(self):
        return "%s [%s]" % (self.value(), self.description())

    def __eq__(self, other):
        if isinstance(other, Transaction):
            return self.value() == other.value() and \
                   self.description() == other.description()
        else:
            return self.value() == _create_decimal(other)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        if isinstance(other, Transaction):
            return self.value() < other.value()
        else:
            return self.value() < _create_decimal(other)

    def __gt__(self, other):
        if isinstance(other, Transaction):
            return self.value() > other.value()
        else:
            return self.value() > _create_decimal(other)

    def __ge__(self, other):
        return self.__eq__(other) or self.__gt__(other)

    def __le__(self, other):
        return self.__eq__(other) or self.__lt__(other)

    def is_null(self):
        """Return whether or not this is a null transaction"""
        return self._value == 0 and self._description is None

    def value(self):
        """Return the value of this transaction. This will be always greater
           than or equal to zero
        """
        return self._value

    def description(self):
        """Return the description of this transaction"""
        return self._description

    @staticmethod
    def maximum_transaction_value():
        """Return the maximum value for a single transaction. Currently this
           is 999999.999999 so that a transaction fits into a f013.6 string
        """
        return _create_decimal(999999.999999)

    @staticmethod
    def round(value):
        """Round the passed floating point value to the precision
           level of the transaction (currently 6 decimal places)
        """
        return _create_decimal(value)

    @staticmethod
    def split(value, description):
        """Split the passed transaction described by 'description' with value
           'value' into a list of transactions that fit the maximum transaction
           limit.
        """
        value = _create_decimal(value)

        if value < Transaction.maximum_transaction_value():
            t = Transaction(value, description)
            return [t]
        else:
            orig_value = value
            values = []

            while value > Transaction.maximum_transaction_value():
                values.append(Transaction.maximum_transaction_value())
                value -= Transaction.maximum_transaction_value()

            if value > 0:
                values.append(value)

            total = 0
            for value in values:
                total += value

            if total != orig_value:
                values[-1] -= (total - orig_value)

            transactions = []

            for i in range(0, len(values)):
                transactions.append(Transaction(values[i], "%s: %d of %d" %
                                    (description, i+1, len(values))))

            values = None

            total = 0
            for transaction in transactions:
                total += transaction.value()

            # ensure that the total is also rounded to 6 dp
            total = _create_decimal(total)

            if total != orig_value:
                from Acquire.Accounting import TransactionError
                raise TransactionError(
                    "Error as split sum (%s) is not equal to the original "
                    "value (%s)" % (total, orig_value))

            return transactions

    @staticmethod
    def from_data(data):
        """Return a newly constructed transaction from the passed dictionary
           that has been decoded from json
        """
        transaction = Transaction()

        if (data and len(data) > 0):
            from Acquire.Accounting import create_decimal as _create_decimal
            transaction._value = _create_decimal(data["value"])
            transaction._description = data["description"]

        return transaction

    def to_data(self):
        """Return this transaction as a dictionary that can be encoded
           to json
        """
        data = {}

        if not self.is_null():
            data["value"] = str(self.value())
            data["description"] = self._description

        return data




import uuid as _uuid
from copy import copy as _copy

from Acquire.Service import Service as _Service

__all__ = ["ComputeService"]


class ComputeService(_Service):
    """This is a specialisation of Service for Compute Services"""
    def __init__(self, other=None):
        if isinstance(other, _Service):
            self.__dict__ = _copy(other.__dict__)

            if not self.is_compute_service():
                from Acquire.Compute import ComputeServiceError
                raise ComputeServiceError(
                    "Cannot construct a ComputeService from "
                    "a service which is not an compute service!")
        else:
            _Service.__init__(self)

    def _call_local_function(self, function, args):
        """Internal function called to short-cut local 'remote'
           function calls
        """
        from compute.route import compute_functions as _compute_functions
        from admin.handler import create_handler as _create_handler
        handler = _create_handler(_compute_functions)
        return handler(function, args)
